"""
Provide the Render API.
"""
from pathlib import Path

from betty.locale import Localizer
from betty.job import Context


class Renderer:
    @property
    def file_extensions(self) -> set[str]:
        raise NotImplementedError(repr(self))

    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: Context | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        raise NotImplementedError(repr(self))


class SequentialRenderer(Renderer):
    def __init__(self, renderers: list[Renderer]):
        self._renderers = renderers
        self._file_extensions = {
            file_extension
            for renderer
            in self._renderers
            for file_extension
            in renderer.file_extensions
        }

    @property
    def file_extensions(self) -> set[str]:
        return self._file_extensions

    async def render_file(
        self,
        file_path: Path,
        *,
        job_context: Context | None = None,
        localizer: Localizer | None = None,
    ) -> Path:
        for renderer in self._renderers:
            if file_path.suffix in renderer.file_extensions:
                return await self.render_file(await renderer.render_file(
                    file_path,
                    job_context=job_context,
                    localizer=localizer,
                ))
        return file_path



"""
Provide the Locale API.
"""
from __future__ import annotations

import asyncio
import calendar
import datetime
import gettext
import glob
import logging
import operator
import threading
from collections import defaultdict
from collections.abc import AsyncIterator
from contextlib import suppress, redirect_stdout, chdir
from functools import total_ordering
from io import StringIO
from pathlib import Path
from typing import Any, Iterator, Sequence, Mapping, Callable, TypeAlias, cast, TYPE_CHECKING
from warnings import warn

import aiofiles
import babel
from aiofiles.os import makedirs
from aiofiles.ospath import exists
from babel import dates, Locale
from babel.messages.frontend import CommandLineInterface
from langcodes import Language
from polib import pofile

from betty import fs
from betty.asyncio import wait
from betty.fs import hashfile, FileSystem, ASSETS_DIRECTORY_PATH, ROOT_DIRECTORY_PATH
from betty.json.linked_data import LinkedDataDumpable, dump_context, add_json_ld
from betty.json.schema import ref_locale, add_property
from betty.serde.dump import DictDump, Dump, dump_default

if TYPE_CHECKING:
    from betty.app import App


DEFAULT_LOCALE = 'en-US'

_LOCALE_DIRECTORY_PATH = ASSETS_DIRECTORY_PATH / 'locale'


class LocaleNotFoundError(RuntimeError):
    def __init__(self, locale: str) -> None:
        super().__init__(f'Cannot find locale "{locale}"')
        self.locale = locale


def to_babel_identifier(locale: Localey) -> str:
    """
    Convert a locale or locale metadata to a Babel locale identifier.
    """
    if isinstance(locale, Locale):
        return str(locale)
    language_data = Language.get(locale)
    return '_'.join(
        part
        for part
        in [
            language_data.language,
            language_data.script,
            language_data.territory,
        ]
        if part
    )


def to_locale(locale: Localey) -> str:
    """
    Ensure that a locale or locale metadata is a locale.
    """
    if isinstance(locale, str):
        return locale
    return '-'.join(
        part
        for part
        in [
            locale.language,
            locale.script,
            locale.territory,
        ]
        if part
    )


Localey: TypeAlias = str | Locale


def get_data(locale: Localey) -> Locale:
    """
    Get locale metadata.
    """
    if isinstance(locale, Locale):
        return locale
    try:
        return Locale.parse(to_babel_identifier(locale))
    except BaseException as e:
        raise LocaleNotFoundError(locale) from e


def get_display_name(locale: Localey, display_locale: Localey | None = None) -> str | None:
    """
    Return a locale's human-readable display name.
    """
    locale_data = get_data(locale)
    return locale_data.get_display_name(
        get_data(display_locale) if display_locale else locale_data
    )


class Localized(LinkedDataDumpable):
    locale: str | None

    def __init__(
        self,
        *args: Any,
        locale: str | None = None,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        self.locale = locale

    async def dump_linked_data(self, app: App) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if self.locale is not None:
            dump['locale'] = self.locale
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        properties = dump_default(schema, 'properties', dict)
        properties['locale'] = ref_locale(schema)
        return schema


class IncompleteDateError(ValueError):
    pass


def _dump_date_iso8601(date: Date) -> str | None:
    if not date.complete:
        return None
    assert date.year
    assert date.month
    assert date.day
    return f'{date.year:04d}-{date.month:02d}-{date.day:02d}'


class Date(LinkedDataDumpable):
    year: int | None
    month: int | None
    day: int | None
    fuzzy: bool

    def __init__(
        self,
        year: int | None = None,
        month: int | None = None,
        day: int | None = None,
        fuzzy: bool = False,
    ):
        self.year = year
        self.month = month
        self.day = day
        self.fuzzy = fuzzy

    def __repr__(self) -> str:
        return '<%s.%s(%s, %s, %s)>' % (self.__class__.__module__, self.__class__.__name__, self.year, self.month, self.day)

    @property
    def comparable(self) -> bool:
        return self.year is not None

    @property
    def complete(self) -> bool:
        return self.year is not None and self.month is not None and self.day is not None

    @property
    def parts(self) -> tuple[int | None, int | None, int | None]:
        return self.year, self.month, self.day

    def to_range(self) -> DateRange:
        if not self.comparable:
            raise ValueError('Cannot convert non-comparable date %s to a date range.' % self)
        if self.month is None:
            month_start = 1
            month_end = 12
        else:
            month_start = month_end = self.month
        if self.day is None:
            day_start = 1
            day_end = calendar.monthrange(
                self.year,  # type: ignore[arg-type]
                month_end,
            )[1]
        else:
            day_start = day_end = self.day
        return DateRange(Date(self.year, month_start, day_start), Date(self.year, month_end, day_end))

    def _compare(self, other: Any, comparator: Callable[[Any, Any], bool]) -> bool:
        if not isinstance(other, Date):
            return NotImplemented
        selfish = self
        if not selfish.comparable or not other.comparable:
            return NotImplemented
        if selfish.complete and other.complete:
            return comparator(selfish.parts, other.parts)
        if not other.complete:
            other = other.to_range()
        if not selfish.complete:
            selfish = selfish.to_range()  # type: ignore[assignment]
        return comparator(selfish, other)

    def __contains__(self, other: Any) -> bool:
        if isinstance(other, Date):
            return self == other
        if isinstance(other, DateRange):
            return self in other
        raise TypeError('Expected to check a %s, but a %s was given' % (type(Datey), type(other)))

    def __lt__(self, other: Any) -> bool:
        return self._compare(other, operator.lt)

    def __le__(self, other: Any) -> bool:
        return self._compare(other, operator.le)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Date):
            return NotImplemented
        return self.parts == other.parts

    def __ge__(self, other: Any) -> bool:
        return self._compare(other, operator.ge)

    def __gt__(self, other: Any) -> bool:
        return self._compare(other, operator.gt)

    async def dump_linked_data(self, app: App, schemas_org: list[str] | None = None) -> DictDump[Dump]:
        dump = await super().dump_linked_data(app)
        if self.year:
            dump['year'] = self.year
        if self.month:
            dump['month'] = self.month
        if self.day:
            dump['day'] = self.day
        if self.comparable:
            dump['iso8601'] = _dump_date_iso8601(self)
        return dump

    async def datey_dump_linked_data(
        self,
        dump: DictDump[Dump],
        start_schema_org: str,
        end_schema_org: str,
    ) -> None:
        if self.comparable:
            dump_context(dump, iso8601=(start_schema_org, end_schema_org))

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        schema['type'] = 'object'
        schema['additionalProperties'] = False
        add_json_ld(schema)
        add_property(schema, 'year', {
            'type': 'number'
        }, False)
        add_property(schema, 'month', {
            'type': 'number'
        }, False)
        add_property(schema, 'day', {
            'type': 'number'
        }, False)
        add_property(schema, 'iso8601', {
            'type': 'string',
            'pattern': '^\\d\\d\\d\\d-\\d\\d-\\d\\d$',
            'description': 'An ISO 8601 date.'
        }, False)
        return schema


async def ref_date(root_schema: DictDump[Dump], app: App) -> DictDump[Dump]:
    """
    Reference the Date schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'date' not in definitions:
        definitions['date'] = await Date.linked_data_schema(app)
    return {
        '$ref': '#/definitions/date',
    }


@total_ordering
class DateRange(LinkedDataDumpable):
    start: Date | None
    start_is_boundary: bool
    end: Date | None
    end_is_boundary: bool

    def __init__(
        self,
        start: Date | None = None,
        end: Date | None = None,
        start_is_boundary: bool = False,
        end_is_boundary: bool = False,
    ):
        self.start = start
        self.start_is_boundary = start_is_boundary
        self.end = end
        self.end_is_boundary = end_is_boundary

    def __repr__(self) -> str:
        return '%s.%s(%s, %s, start_is_boundary=%s, end_is_boundary=%s)' % (self.__class__.__module__, self.__class__.__name__, repr(self.start), repr(self.end), repr(self.start_is_boundary), repr(self.end_is_boundary))

    @property
    def comparable(self) -> bool:
        return self.start is not None and self.start.comparable or self.end is not None and self.end.comparable

    def __contains__(self, other: Any) -> bool:
        if not self.comparable:
            return False

        if isinstance(other, Date):
            others = [other]
        elif isinstance(other, DateRange):
            if not other.comparable:
                return False
            others = []
            if other.start is not None and other.start.comparable:
                others.append(other.start)
            if other.end is not None and other.end.comparable:
                others.append(other.end)
        else:
            raise TypeError('Expected to check a %s, but a %s was given' % (type(Datey), type(other)))

        if self.start is not None and self.end is not None:
            if isinstance(other, DateRange) and (other.start is None or other.end is None):
                if other.start is None:
                    return self.start <= other.end or self.end <= other.end
                if other.end is None:
                    return self.start >= other.start or self.end >= other.start
            for another in others:
                if self.start <= another <= self.end:
                    return True
            if isinstance(other, DateRange):
                for selfdate in [self.start, self.end]:
                    if other.start <= selfdate <= other.end:
                        return True

        elif self.start is not None:
            # Two date ranges with start dates only always overlap.
            if isinstance(other, DateRange) and other.end is None:
                return True

            for other in others:
                if self.start <= other:
                    return True
        elif self.end is not None:
            # Two date ranges with end dates only always overlap.
            if isinstance(other, DateRange) and other.start is None:
                return True

            for other in others:
                if other <= self.end:
                    return True
        return False

    async def dump_linked_data(
        self,
        app: App,
        start_schema_org: str | None = None,
        end_schema_org: str | None = None,
    ) -> DictDump[Dump]:
        dump: DictDump[Dump] = {}
        if self.start:
            dump['start'] = await self.start.dump_linked_data(
                app,
                [start_schema_org] if start_schema_org else None,
            )
        if self.end:
            dump['end'] = await self.end.dump_linked_data(
                app,
                [end_schema_org] if end_schema_org else None,
            )
        return dump

    @classmethod
    async def linked_data_schema(cls, app: App) -> DictDump[Dump]:
        schema = await super().linked_data_schema(app)
        schema['type'] = 'object'
        schema['additionalProperties'] = False
        add_property(schema, 'start', await ref_date(schema, app), False)
        add_property(schema, 'end', await ref_date(schema, app), False)
        return schema

    async def datey_dump_linked_data(
        self,
        dump: DictDump[Dump],
        start_schema_org: str,
        end_schema_org: str,
    ) -> None:
        if self.start and self.start.comparable:
            start = dump_default(dump, 'start', dict)
            dump_context(start, iso8601=start_schema_org)
        if self.end and self.end.comparable:
            end = dump_default(dump, 'end', dict)
            dump_context(end, iso8601=end_schema_org)

    def _get_comparable_date(self, date: Date | None) -> Date | None:
        if date and date.comparable:
            return date
        return None

    _LT_DATE_RANGE_COMPARATORS = {
        (True, True, True, True): lambda self_start, self_end, other_start, other_end: self_start < other_start,
        (True, True, True, False): lambda self_start, self_end, other_start, other_end: self_start <= other_start,
        (True, True, False, True): lambda self_start, self_end, other_start, other_end: self_start < other_end or self_end <= other_end,
        (True, True, False, False): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (True, False, True, True): lambda self_start, self_end, other_start, other_end: self_start < other_start,
        (True, False, True, False): lambda self_start, self_end, other_start, other_end: self_start < other_start,
        (True, False, False, True): lambda self_start, self_end, other_start, other_end: self_start < other_end,
        (True, False, False, False): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (False, True, True, True): lambda self_start, self_end, other_start, other_end: self_end <= other_start,
        (False, True, True, False): lambda self_start, self_end, other_start, other_end: self_end <= other_start,
        (False, True, False, True): lambda self_start, self_end, other_start, other_end: self_end < other_end,
        (False, True, False, False): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (False, False, True, True): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (False, False, True, False): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (False, False, False, True): lambda self_start, self_end, other_start, other_end: NotImplemented,
        (False, False, False, False): lambda self_start, self_end, other_start, other_end: NotImplemented,
    }

    _LT_DATE_COMPARATORS = {
        (True, True): lambda self_start, self_end, other: self_start < other,
        (True, False): lambda self_start, self_end, other: self_start < other,
        (False, True): lambda self_start, self_end, other: self_end <= other,
        (False, False): lambda self_start, self_end, other: NotImplemented,
    }

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (Date, DateRange)):
            return NotImplemented

        self_start = self._get_comparable_date(self.start)
        self_end = self._get_comparable_date(self.end)
        signature = (
            self_start is not None,
            self_end is not None,
        )
        if isinstance(other, DateRange):
            other_start = self._get_comparable_date(other.start)
            other_end = self._get_comparable_date(other.end)
            return self._LT_DATE_RANGE_COMPARATORS[(
                *signature,
                other_start is not None,
                other_end is not None,
            )](self_start, self_end, other_start, other_end)
        else:
            return self._LT_DATE_COMPARATORS[signature](self_start, self_end, other)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, Date):
            return False

        if not isinstance(other, DateRange):
            return NotImplemented
        return (self.start, self.end, self.start_is_boundary, self.end_is_boundary) == (other.start, other.end, other.start_is_boundary, other.end_is_boundary)


async def ref_date_range(root_schema: DictDump[Dump], app: App) -> DictDump[Dump]:
    """
    Reference the DateRange schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'dateRange' not in definitions:
        definitions['dateRange'] = await DateRange.linked_data_schema(app)
    return {
        '$ref': '#/definitions/dateRange',
    }


async def ref_datey(root_schema: DictDump[Dump], app: App) -> DictDump[Dump]:
    """
    Reference the Datey schema.
    """
    definitions = dump_default(root_schema, 'definitions', dict)
    if 'datey' not in definitions:
        definitions['datey'] = {
            'oneOf': [
                await ref_date(root_schema, app),
                await ref_date_range(root_schema, app),
            ]
        }
    return {
        '$ref': '#/definitions/datey',
    }


Datey: TypeAlias = Date | DateRange
DatePartsFormatters: TypeAlias = Mapping[tuple[bool, bool, bool], str]
DateFormatters: TypeAlias = Mapping[tuple[bool | None], str]
DateRangeFormatters: TypeAlias = Mapping[tuple[bool | None, bool | None, bool | None, bool | None], str]


class Localizer:
    def __init__(self, locale: str, translations: gettext.NullTranslations):
        self._locale = locale
        self._locale_data = get_data(locale)
        self._translations = translations
        self.__date_parts_formatters: DatePartsFormatters | None = None
        self.__date_formatters: DateFormatters | None = None
        self.__date_range_formatters: DateRangeFormatters | None = None

    @property
    def locale(self) -> str:
        return self._locale

    def _(self, message: str) -> str:
        return self._translations.gettext(message)

    def gettext(self, message: str) -> str:
        return self._translations.gettext(message)

    def ngettext(self, message_singular: str, message_plural: str, n: int) -> str:
        return self._translations.ngettext(message_singular, message_plural, n)

    def pgettext(self, context: str, message: str) -> str:
        return self._translations.pgettext(context, message)

    def npgettext(self, context: str, message_singular: str, message_plural: str, n: int) -> str:
        return self._translations.npgettext(context, message_singular, message_plural, n)

    @property
    def _date_parts_formatters(self) -> DatePartsFormatters:
        if self.__date_parts_formatters is None:
            self.__date_parts_formatters = {
                (True, True, True): self._('MMMM d, y'),
                (True, True, False): self._('MMMM, y'),
                (True, False, False): self._('y'),
                (False, True, True): self._('MMMM d'),
                (False, True, False): self._('MMMM'),
            }
        return self.__date_parts_formatters

    @property
    def _date_formatters(self) -> DateFormatters:
        if self.__date_formatters is None:
            self.__date_formatters = {
                (True,): self._('around {date}'),
                (False,): self._('{date}'),
            }
        return self.__date_formatters

    @property
    def _date_range_formatters(self) -> DateRangeFormatters:
        if self.__date_range_formatters is None:
            self.__date_range_formatters = {
                (False, False, False, False): self._('from {start_date} until {end_date}'),
                (False, False, False, True): self._('from {start_date} until sometime before {end_date}'),
                (False, False, True, False): self._('from {start_date} until around {end_date}'),
                (False, False, True, True): self._('from {start_date} until sometime before around {end_date}'),
                (False, True, False, False): self._('from sometime after {start_date} until {end_date}'),
                (False, True, False, True): self._('sometime between {start_date} and {end_date}'),
                (False, True, True, False): self._('from sometime after {start_date} until around {end_date}'),
                (False, True, True, True): self._('sometime between {start_date} and around {end_date}'),
                (True, False, False, False): self._('from around {start_date} until {end_date}'),
                (True, False, False, True): self._('from around {start_date} until sometime before {end_date}'),
                (True, False, True, False): self._('from around {start_date} until around {end_date}'),
                (True, False, True, True): self._('from around {start_date} until sometime before around {end_date}'),
                (True, True, False, False): self._('from sometime after around {start_date} until {end_date}'),
                (True, True, False, True): self._('sometime between around {start_date} and {end_date}'),
                (True, True, True, False): self._('from sometime after around {start_date} until around {end_date}'),
                (True, True, True, True): self._('sometime between around {start_date} and around {end_date}'),
                (False, False, None, None): self._('from {start_date}'),
                (False, True, None, None): self._('sometime after {start_date}'),
                (True, False, None, None): self._('from around {start_date}'),
                (True, True, None, None): self._('sometime after around {start_date}'),
                (None, None, False, False): self._('until {end_date}'),
                (None, None, False, True): self. _('sometime before {end_date}'),
                (None, None, True, False): self._('until around {end_date}'),
                (None, None, True, True): self._('sometime before around {end_date}'),
            }
        return self.__date_range_formatters

    def format_datey(self, date: Datey) -> str:
        """
        Format a datey value into a human-readable string.
        """
        if isinstance(date, Date):
            return self.format_date(date)
        return self.format_date_range(date)

    def format_date(self, date: Date) -> str:
        try:
            return self._date_formatters[(date.fuzzy,)].format(
                date=self._format_date_parts(date),
            )
        except IncompleteDateError:
            return self._('unknown date')

    def _format_date_parts(self, date: Date | None) -> str:
        if date is None:
            raise IncompleteDateError('This date is None.')
        try:
            date_parts_format = self._date_parts_formatters[tuple(
                map(lambda x: x is not None, date.parts),  # type: ignore[index]
            )]
        except KeyError:
            raise IncompleteDateError('This date does not have enough parts to be rendered.')
        parts = map(lambda x: 1 if x is None else x, date.parts)
        return dates.format_date(datetime.date(*parts), date_parts_format, self._locale_data)

    def format_date_range(self, date_range: DateRange) -> str:
        formatter_configuration: tuple[bool | None, bool | None, bool | None, bool | None] = (None, None, None, None)
        formatter_arguments = {}

        with suppress(IncompleteDateError):
            formatter_arguments['start_date'] = self._format_date_parts(date_range.start)
            formatter_configuration = (
                None if date_range.start is None else date_range.start.fuzzy,
                date_range.start_is_boundary,
                formatter_configuration[2],
                formatter_configuration[3],
            )

        with suppress(IncompleteDateError):
            formatter_arguments['end_date'] = self._format_date_parts(date_range.end)
            formatter_configuration = (
                formatter_configuration[0],
                formatter_configuration[1],
                None if date_range.end is None else date_range.end.fuzzy,
                date_range.end_is_boundary,
            )

        if not formatter_arguments:
            raise IncompleteDateError('This date range does not have enough parts to be rendered.')

        return self._date_range_formatters[formatter_configuration].format(**formatter_arguments)


DEFAULT_LOCALIZER = Localizer(DEFAULT_LOCALE, gettext.NullTranslations())


class LocalizerRepository:
    def __init__(self, assets: FileSystem):
        self._assets = assets
        self._localizers: dict[str, Localizer] = {}
        self._locks: dict[str, threading.Lock] = defaultdict(threading.Lock)

    @property
    def locales(self) -> Iterator[str]:
        yield DEFAULT_LOCALE
        for assets_directory_path, __ in reversed(self._assets.paths):
            for po_file_path in glob.glob(str(assets_directory_path / 'locale' / '*' / 'betty.po')):
                yield Path(po_file_path).parent.name

    async def get(self, locale: Localey) -> Localizer:
        locale = to_locale(locale)
        try:
            return self._localizers[locale]
        except KeyError:
            return await self._build_translation(locale)

    def __getitem__(self, locale: Localey) -> Localizer:
        locale = to_locale(locale)
        try:
            return self._localizers[locale]
        except KeyError:
            return wait(self._build_translation(locale))

    async def get_negotiated(self, *preferred_locales: str) -> Localizer:
        preferred_locales = (*preferred_locales, DEFAULT_LOCALE)
        negotiated_locale = negotiate_locale(
            preferred_locales,
            [
                str(get_data(locale))
                for locale
                in self.locales
            ],
        )
        return await self.get(negotiated_locale or DEFAULT_LOCALE)

    async def _build_translation(self, locale: str) -> Localizer:
        translations = gettext.NullTranslations()
        for assets_directory_path, __ in reversed(self._assets.paths):
            opened_translations = await asyncio.to_thread(self._open_translations, locale, assets_directory_path)
            if opened_translations:
                opened_translations.add_fallback(translations)
                translations = opened_translations
        self._localizers[locale] = Localizer(locale, translations)
        return self._localizers[locale]

    def _open_translations(self, locale: str, assets_directory_path: Path) -> gettext.GNUTranslations | None:
        po_file_path = assets_directory_path / 'locale' / locale / 'betty.po'
        try:
            translation_version = hashfile(po_file_path)
        except FileNotFoundError:
            return None
        cache_directory_path = fs.CACHE_DIRECTORY_PATH / 'locale' / translation_version
        mo_file_path = cache_directory_path / 'betty.mo'

        with self._locks[locale]:
            with suppress(FileNotFoundError):
                with open(mo_file_path, 'rb') as f:
                    return gettext.GNUTranslations(f)

            cache_directory_path.mkdir(exist_ok=True, parents=True)

            with redirect_stdout(StringIO()):
                CommandLineInterface().run([
                    '',
                    'compile',
                    '-i',
                    str(po_file_path),
                    '-o',
                    str(mo_file_path),
                    '-l',
                    str(get_data(locale)),
                    '-D',
                    'betty',
                ])
        with open(mo_file_path, 'rb') as f:
            return gettext.GNUTranslations(f)

    async def coverage(self, locale: Localey) -> tuple[int, int]:
        translatables = {translatable async for translatable in self._get_translatables()}
        locale = to_locale(locale)
        if locale == DEFAULT_LOCALE:
            return len(translatables), len(translatables)
        translations = {translation async for translation in self._get_translations(locale)}
        return len(translations), len(translatables)

    async def _get_translatables(self) -> AsyncIterator[str]:
        for assets_directory_path, __ in self._assets.paths:
            with suppress(FileNotFoundError):
                async with aiofiles.open(assets_directory_path / 'betty.pot') as pot_data_f:
                    pot_data = await pot_data_f.read()
                    for entry in pofile(pot_data):
                        yield entry.msgid_with_context

    async def _get_translations(self, locale: str) -> AsyncIterator[str]:
        for assets_directory_path, __ in reversed(self._assets.paths):
            with suppress(FileNotFoundError):
                async with aiofiles.open(
                    assets_directory_path / 'locale' / locale / 'betty.po',
                    encoding='utf-8',
                ) as p_data_f:
                    po_data = await p_data_f.read()
                    for entry in pofile(po_data):
                        if entry.translated():
                            yield entry.msgid_with_context


class Localizable:
    def localize(self, localizer: Localizer) -> str:
        raise NotImplementedError

    def __str__(self) -> str:
        localized = self.localize(DEFAULT_LOCALIZER)
        warn(f'{type(self)} ("{localized}") SHOULD NOT be cast to a string. Instead, call {type(self)}.localize() to ensure it is always formatted in the desired locale.')
        return localized


class Str(Localizable):
    def _localize_format_kwargs(self, localizer: Localizer, **format_kwargs: str | Localizable) -> dict[str, str]:
        return {
            key: value.localize(localizer) if isinstance(value, Localizable) else value
            for key, value
            in format_kwargs.items()
        }

    @classmethod
    def plain(cls, plain: Any, **format_kwargs: str | Localizable) -> Str:
        return _PlainStr(str(plain), **format_kwargs)

    @classmethod
    def call(cls, call: Callable[[Localizer], str]) -> Str:
        return _CallStr(call)

    @classmethod
    def _(cls, message: str, **format_kwargs: str | Localizable) -> Str:
        return cls.gettext(message, **format_kwargs)

    @classmethod
    def gettext(cls, message: str, **format_kwargs: str | Localizable) -> Str:
        return _GettextStr('gettext', message, **format_kwargs)

    @classmethod
    def ngettext(cls, message_singular: str, message_plural: str, n: int, **format_kwargs: str | Localizable) -> Str:
        return _GettextStr('ngettext', message_singular, message_plural, n, **format_kwargs)

    @classmethod
    def pgettext(cls, context: str, message: str, **format_kwargs: str | Localizable) -> Str:
        return _GettextStr('pgettext', context, message, **format_kwargs)

    @classmethod
    def npgettext(cls, context: str, message_singular: str, message_plural: str, n: int, **format_kwargs: str | Localizable) -> Str:
        return _GettextStr('npgettext', context, message_singular, message_plural, n, **format_kwargs)


class _PlainStr(Str):
    def __init__(self, plain: str, **format_kwargs: str | Localizable):
        self._plain = plain
        self._format_kwargs = format_kwargs

    def localize(self, localizer: Localizer) -> str:
        return self._plain.format(**self._localize_format_kwargs(localizer, **self._format_kwargs))


class _CallStr(Str):
    def __init__(self, call: Callable[[Localizer], str]):
        self._call = call

    def localize(self, localizer: Localizer) -> str:
        return self._call(localizer)


class _GettextStr(Str):
    def __init__(self, gettext_method_name: str, *gettext_args: Any, **format_kwargs: str | Localizable) -> None:
        self._gettext_method_name = gettext_method_name
        self._gettext_args = gettext_args
        self._format_kwargs = format_kwargs

    def localize(self, localizer: Localizer) -> str:
        return cast(str, getattr(localizer, self._gettext_method_name)(*self._gettext_args)).format(**self._localize_format_kwargs(localizer, **self._format_kwargs))


def negotiate_locale(preferred_locales: Localey | Sequence[Localey], available_locales: Sequence[Localey]) -> Locale | None:
    """
    Negotiate the preferred locale from a sequence.
    """
    if isinstance(preferred_locales, (str, Locale)):
        preferred_locales = [preferred_locales]
    return _negotiate_locale(
        [
            to_babel_identifier(locale)
            for locale
            in preferred_locales
        ],
        {
            to_babel_identifier(locale)
            for locale
            in available_locales
        },
        False,
    )


def _negotiate_locale(preferred_locale_babel_identifiers: Sequence[str], available_locale_babel_identifiers: set[str], root: bool) -> Locale | None:
    negotiated_locale = babel.negotiate_locale(preferred_locale_babel_identifiers, available_locale_babel_identifiers)
    if negotiated_locale is not None:
        return Locale.parse(negotiated_locale)
    if not root:
        return _negotiate_locale(
            [
                babel_identifier.split('_')[0] if '_' in babel_identifier else babel_identifier
                for babel_identifier
                in preferred_locale_babel_identifiers
            ],
            available_locale_babel_identifiers,
            True,
        )
    return None


def negotiate_localizeds(preferred_locales: Localey | Sequence[Localey], localizeds: Sequence[Localized]) -> Localized | None:
    """
    Negotiate the preferred localized value from a sequence.
    """
    negotiated_locale_data = negotiate_locale(
        preferred_locales,
        [
            localized.locale
            for localized
            in localizeds
            if localized.locale is not None
        ],
    )
    if negotiated_locale_data is not None:
        negotiated_locale = to_locale(negotiated_locale_data)
        for localized in localizeds:
            if localized.locale == negotiated_locale:
                return localized
    for localized in localizeds:
        if localized.locale is None:
            return localized
    with suppress(IndexError):
        return localizeds[0]
    return None


async def init_translation(locale: str) -> None:
    """
    Initialize a new translation.
    """
    po_file_path = _LOCALE_DIRECTORY_PATH / locale / 'betty.po'
    with redirect_stdout(StringIO()):
        if await exists(po_file_path):
            logging.getLogger(__name__).info(f'Translations for {locale} already exist at {po_file_path}.')
            return

        locale_data = get_data(locale)
        CommandLineInterface().run([
            '',
            'init',
            '--no-wrap',
            '-i',
            str(ASSETS_DIRECTORY_PATH / 'betty.pot'),
            '-o',
            str(po_file_path),
            '-l',
            str(locale_data),
            '-D',
            'betty',
        ])
        logging.getLogger(__name__).info(f'Translations for {locale} initialized at {po_file_path}.')


async def update_translations(_output_assets_directory_path: Path = ASSETS_DIRECTORY_PATH) -> None:
    """
    Update all existing translations based on changes in translatable strings.
    """
    source_paths = glob.glob('betty/*')
    # Remove the tests directory from the extraction,
    # or we'll be seeing some unusual additions to the translations.
    source_paths.remove(str(Path('betty') / 'tests'))
    pot_file_path = _output_assets_directory_path / 'betty.pot'
    with redirect_stdout(StringIO()):
        with chdir(ROOT_DIRECTORY_PATH):
            CommandLineInterface().run([
                '',
                'extract',
                '--no-location',
                '--no-wrap',
                '--sort-output',
                '-F',
                'babel.ini',
                '-o',
                str(pot_file_path),
                '--project',
                'Betty',
                '--copyright-holder',
                'Bart Feenstra & contributors',
                *source_paths,
            ])
            for po_file_path_str in glob.glob('betty/assets/locale/*/betty.po'):
                po_file_path = (_output_assets_directory_path / (ROOT_DIRECTORY_PATH / po_file_path_str).relative_to(ASSETS_DIRECTORY_PATH)).resolve()
                await makedirs(po_file_path.parent, exist_ok=True)
                po_file_path.touch()
                locale = po_file_path.parent.name
                locale_data = get_data(locale)
                CommandLineInterface().run([
                    '',
                    'update',
                    '-i',
                    str(pot_file_path),
                    '-o',
                    str(po_file_path),
                    '-l',
                    str(locale_data),
                    '-D',
                    'betty',
                ])



"""
Provide utilities for concurrent programming.
"""
import asyncio
import time
from asyncio import sleep
from math import floor
from threading import Lock
from types import TracebackType


class AsynchronizedLock:
    """
    Make a sychronous (blocking) lock asynchronous (non-blocking).
    """

    def __init__(self, lock: Lock):
        self._lock = lock

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        self.release()

    async def acquire(self) -> None:
        while not self._lock.acquire(False):
            # Sleeping for zero seconds does not actually sleep, but gives the event
            # loop a chance to progress other tasks while we wait for another chance
            # to acquire the lock.
            await sleep(0)

    def release(self) -> None:
        self._lock.release()


class RateLimiter:
    """
    Rate-limit operations.

    This class implements the `Token Bucket algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_.
    """

    _PERIOD = 1

    def __init__(self, maximum: int):
        self._lock = AsynchronizedLock(Lock())
        self._maximum = maximum
        self._available: int | float = maximum
        # A Token Bucket fills as time passes. However, we want callers to be able to start
        # using the limiter immediately, so we 'preload' the first's period's tokens, and
        # set the last added time to the end of the first period. This ensures there is no
        # needless waiting if the number of tokens consumed in total is less than the limit
        # per period.
        self._last_add = time.monotonic() + self._PERIOD

    def _add_tokens(self):
        now = time.monotonic()
        elapsed = now - self._last_add
        added = elapsed * self._maximum
        possibly_available = floor(self._available + added)
        if possibly_available > 0:
            self._available = min(possibly_available, self._maximum)
            self._last_add = now

    async def __aenter__(self) -> None:
        await self.wait()

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        return

    async def wait(self) -> None:
        async with self._lock:
            while self._available < 1:
                self._add_tokens()
                if self._available < 1:
                    await asyncio.sleep(0.1)
            self._available -= 1



"""
Provide the Serve API to serve resources within the application.
"""
from __future__ import annotations

import contextlib
import logging
import threading
import webbrowser
from http.server import SimpleHTTPRequestHandler, HTTPServer
from io import StringIO
from pathlib import Path
from types import TracebackType
from typing import Sequence, Any

from aiofiles.os import makedirs, symlink
from aiofiles.tempfile import TemporaryDirectory, AiofilesContextManagerTempDir
from aiohttp import ClientSession

from betty.app import App
from betty.error import UserFacingError
from betty.functools import Do
from betty.locale import Str, Localizer

DEFAULT_PORT = 8000


class ServerNotStartedError(RuntimeError):
    pass


class NoPublicUrlBecauseServerNotStartedError(ServerNotStartedError):
    def __init__(self):
        super().__init__('Cannot get the public URL for a server that has not started yet.')


class OsError(UserFacingError, OSError):
    pass


class Server:
    """
    Provide a development web server.
    """

    def __init__(self, localizer: Localizer):
        self._localizer = localizer

    @classmethod
    def name(cls) -> str:
        return f'{cls.__module__}.{cls.__name__}'

    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    async def start(self) -> None:
        """
        Start the server.
        """
        pass

    async def show(self) -> None:
        """
        Show the served site to the user.
        """
        logging.getLogger(__name__).info(self._localizer._('Serving your site at {url}...').format(
            url=self.public_url,
        ))
        webbrowser.open_new_tab(self.public_url)

    async def stop(self) -> None:
        """
        Stop the server.
        """
        pass

    @property
    def public_url(self) -> str:
        raise NotImplementedError(repr(self))

    async def __aenter__(self) -> Server:
        await self.start()
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        await self.stop()

    async def assert_available(self) -> None:
        # @todo In Betty 0.4.0, require the app's existing client session.
        async with ClientSession() as session:
            try:
                await Do[Any, None](self._assert_available, session).until()
            except BaseException:
                raise UserFacingError(Str._('The server was unreachable after starting.'))

    async def _assert_available(self, session: ClientSession) -> None:
        async with session.get(self.public_url) as response:
            assert response.status == 200


class AppServer(Server):
    def __init__(self, app: App) -> None:
        super().__init__(localizer=app.localizer)
        self._app = app

    @staticmethod
    def get(app: App) -> AppServer:
        for server in app.servers.values():
            if isinstance(server, AppServer):
                return server
        raise RuntimeError(f'Cannot find a project server. This must never happen, because {BuiltinAppServer} should be the fallback.')

    async def start(self) -> None:
        await makedirs(self._app.project.configuration.www_directory_path, exist_ok=True)
        await super().start()


class ServerProvider:
    @property
    def servers(self) -> Sequence[Server]:
        raise NotImplementedError(repr(self))


class _BuiltinServerRequestHandler(SimpleHTTPRequestHandler):
    def end_headers(self) -> None:
        self.send_header('Cache-Control', 'no-cache')
        super().end_headers()


class BuiltinServer(Server):
    def __init__(
        self,
        www_directory_path: Path,
        *,
        root_path: str | None = None,
        localizer: Localizer,
    ) -> None:
        super().__init__(localizer)
        self._www_directory_path = www_directory_path
        self._root_path = root_path
        self._http_server: HTTPServer | None = None
        self._port: int | None = None
        self._thread: threading.Thread | None = None
        self._temporary_root_directory: AiofilesContextManagerTempDir[None, Any, Any] | None = None

    @classmethod
    def label(cls) -> Str:
        return Str._('Python built-in')

    async def start(self) -> None:
        await super().start()
        if self._root_path:
            # To mimic the root path, symlink the project's WWW directory into a temporary
            # directory, so we do not have to make changes to any existing files.
            self._temporary_root_directory = TemporaryDirectory()
            temporary_root_directory_path = Path(await self._temporary_root_directory.__aenter__())
            temporary_www_directory = temporary_root_directory_path
            for root_path_component in self._root_path.split('/'):
                temporary_www_directory /= root_path_component
            if temporary_www_directory != temporary_root_directory_path:
                await symlink(self._www_directory_path, temporary_www_directory)
            www_directory_path = temporary_root_directory_path
        else:
            www_directory_path = self._www_directory_path
        logging.getLogger(__name__).info(self._localizer._("Starting Python's built-in web server..."))
        for self._port in range(DEFAULT_PORT, 65535):
            with contextlib.suppress(OSError):
                self._http_server = HTTPServer(
                    ('', self._port),
                    lambda request, client_address, server: _BuiltinServerRequestHandler(
                        request,
                        client_address,
                        server,
                        directory=str(www_directory_path),
                    ),
                )
                break
        if self._http_server is None:
            raise OsError(Str._('Cannot find an available port to bind the web server to.'))
        self._thread = threading.Thread(target=self._serve)
        self._thread.start()
        await self.assert_available()

    @property
    def public_url(self) -> str:
        if self._port is not None:
            url = f'http://localhost:{self._port}'
            if self._root_path:
                url = f'{url}/{self._root_path}'
            return url
        raise NoPublicUrlBecauseServerNotStartedError()

    def _serve(self) -> None:
        with contextlib.redirect_stderr(StringIO()):
            assert self._http_server
            self._http_server.serve_forever()

    async def stop(self) -> None:
        await super().stop()
        if self._http_server is not None:
            self._http_server.shutdown()
            self._http_server.server_close()
        if self._thread is not None:
            self._thread.join()
        if self._temporary_root_directory is not None:
            await self._temporary_root_directory.__aexit__(None, None, None)


class BuiltinAppServer(AppServer):
    def __init__(self, app: App) -> None:
        super().__init__(app)
        self._server = BuiltinServer(
            self._app.project.configuration.www_directory_path,
            root_path=self._app.project.configuration.root_path,
            localizer=self._app.localizer
        )

    @classmethod
    def label(cls) -> Str:
        return BuiltinServer.label()

    @property
    def public_url(self) -> str:
        return self._server.public_url

    async def start(self) -> None:
        await super().start()
        await self._server.start()

    async def stop(self) -> None:
        await super().stop()
        await self._server.stop()



"""
Provide Betty's default Jinja2 filters.
"""
from __future__ import annotations

import json as stdjson
import re
import warnings
from asyncio import get_running_loop
from base64 import b64encode
from collections.abc import Awaitable
from contextlib import suppress
from io import BytesIO
from pathlib import Path
from typing import Callable, Iterable, Any, Iterator, TypeVar, AsyncIterator
from urllib.parse import quote

import aiofiles
from PIL import Image
from PIL.Image import DecompressionBombWarning
from aiofiles.os import makedirs
from geopy import units
from geopy.format import DEGREES_FORMAT
from jinja2 import pass_context, \
    pass_eval_context
from jinja2.async_utils import auto_aiter, auto_await
from jinja2.filters import prepare_map, make_attrgetter
from jinja2.nodes import EvalContext
from jinja2.runtime import Context, Macro
from markupsafe import Markup, escape
from pdf2image.pdf2image import convert_from_path

from betty import _resizeimage
from betty.asyncio import sync
from betty.fs import hashfile
from betty.functools import walk
from betty.locale import negotiate_localizeds, Localized, Datey, negotiate_locale, Localey, get_data, Localizable
from betty.media_type import MediaType
from betty.model import get_entity_type_name
from betty.model.ancestry import File, Dated
from betty.os import link_or_copy
from betty.serde.dump import minimize, none_void, void_none
from betty.string import camel_case_to_snake_case, camel_case_to_kebab_case, upper_camel_case_to_lower_camel_case

T = TypeVar('T')


@pass_context
def filter_url(
    context: Context,
    resource: Any,
    media_type: str | None = None,
    *args: Any,
    locale: Localey | None = None,
    **kwargs: Any,
) -> str:
    """
    Generate a localized URL for a localizable resource.
    """
    from betty.jinja2 import context_app, context_localizer

    return context_app(context).url_generator.generate(
        resource,
        media_type or 'text/html',
        *args,
        locale=locale or context_localizer(context).locale,  # type: ignore[misc]
        **kwargs,
    )


@pass_context
def filter_static_url(
    context: Context,
    resource: Any,
    absolute: bool = False,
) -> str:
    """
    Generate a static URL for a static resource.
    """
    from betty.jinja2 import context_app

    return context_app(context).static_url_generator.generate(
        resource,
        absolute=absolute,
    )


@pass_context
def filter_localize(
    context: Context,
    localizable: Localizable,
) -> str:
    """
    Localize a value using the context's current localizer.
    """
    from betty.jinja2 import context_localizer

    return localizable.localize(context_localizer(context))


@pass_context
def filter_format_datey(
    context: Context,
    datey: Datey,
) -> str:
    """
    Format a date or a date range.
    """
    from betty.jinja2 import context_localizer

    return context_localizer(context).format_datey(datey)


def filter_json(data: Any, indent: int | None = None) -> str:
    """
    Convert a value to a JSON string.
    """
    return stdjson.dumps(data, indent=indent)


async def filter_flatten(values_of_values: Iterable[Iterable[T]]) -> AsyncIterator[T]:
    """
    Flatten an iterable of iterables into a single iterable.
    """
    async for values in auto_aiter(values_of_values):
        async for value in auto_aiter(values):
            yield value


def filter_walk(value: Any, attribute_name: str) -> Iterable[Any]:
    """
    Walk over a data structure.
    """
    return walk(value, attribute_name)


_paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')


@pass_eval_context
def filter_paragraphs(eval_ctx: EvalContext, text: str) -> str | Markup:
    """
    Convert newlines to <p> and <br> tags.

    Taken from http://jinja.pocoo.org/docs/2.10/api/#custom-filters.
    """
    result = u'\n\n'.join(u'<p>%s</p>' % p.replace('\n', Markup('<br>\n'))
                          for p in _paragraph_re.split(escape(text)))
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def filter_format_degrees(degrees: int) -> str:
    """
    Format geographic coordinates.
    """
    arcminutes = units.arcminutes(degrees=degrees - int(degrees))
    arcseconds = units.arcseconds(arcminutes=arcminutes - int(arcminutes))
    format_dict = dict(
        deg='°',
        arcmin="'",
        arcsec='"',
        degrees=degrees,
        minutes=round(abs(arcminutes)),
        seconds=round(abs(arcseconds))
    )
    return DEGREES_FORMAT % format_dict  # type: ignore[no-any-return]


async def filter_unique(value: Iterable[T]) -> AsyncIterator[T]:
    """
    Iterate over an iterable of values and only yield those values that have not been yielded before.
    """
    seen = []
    async for value in auto_aiter(value):
        if value not in seen:
            yield value
            seen.append(value)


@pass_context
async def filter_map(context: Context, values: Iterable[Any], *args: Any, **kwargs: Any) -> Any:
    """
    Map an iterable's values.

    This mimics Jinja2's built-in map filter, but allows macros as callbacks.
    """
    if len(args) > 0 and isinstance(args[0], Macro):
        func: Macro | Callable[[Any], bool] = args[0]
    else:
        func = prepare_map(context, args, kwargs)
    async for value in auto_aiter(values):
        yield await auto_await(func(value))


@pass_context
async def filter_file(context: Context, file: File) -> str:
    """
    Preprocess a file for use in a page.

    :return: The public path to the preprocessed file. This can be used on a web page.
    """
    from betty.jinja2 import context_app, context_job_context

    app = context_app(context)
    job_context = context_job_context(context)

    if job_context is None or job_context.claim(f'filter_file:{file.id}'):
        file_destination_path = app.project.configuration.www_directory_path / 'file' / file.id / 'file' / file.path.name
        await makedirs(file_destination_path.parent, exist_ok=True)
        await link_or_copy(file.path, file_destination_path)

    return f'/file/{quote(file.id)}/file/{quote(file.path.name)}'


@pass_context
async def filter_image(
    context: Context,
    file: File,
    width: int | None = None,
    height: int | None = None,
) -> str:
    """
    Preprocess an image file for use in a page.

    :return: The public path to the preprocessed file. This can be embedded in a web page.
    """
    from betty.jinja2 import context_app, context_job_context

    # Treat SVGs as regular files.
    if file.media_type and file.media_type.type == 'image' and 'svg+xml' == file.media_type.subtype:
        return await filter_file(context, file)

    app = context_app(context)
    job_context = context_job_context(context)

    destination_name = f'{file.id}-'
    if height and width:
        destination_name += f'{width}x{height}'
    elif height:
        destination_name += f'-x{height}'
    elif width:
        destination_name += f'{width}x-'
    else:
        raise ValueError('At least the width or height must be given.')

    file_directory_path = app.project.configuration.www_directory_path / 'file'

    if file.media_type:
        if file.media_type.type == 'image':
            image_loader = _load_image_image
            destination_name += file.path.suffix
        elif file.media_type.type == 'application' and file.media_type.subtype == 'pdf':
            image_loader = _load_image_application_pdf
            destination_name += '.' + 'jpg'
        else:
            raise ValueError(f'Cannot convert a file of media type "{file.media_type}" to an image.')
    else:
        raise ValueError('Cannot convert a file without a media type to an image.')

    cache_item_id = f'filter_image:{hashfile(file.path)}:{"" if width is None else width}:{"" if height is None else height}'
    if job_context is None or job_context.claim(cache_item_id):
        loop = get_running_loop()
        await loop.run_in_executor(
            app.process_pool,
            _execute_filter_image,
            image_loader,
            file.path,
            file.media_type,
            app.cache.path / 'image' / filter_base64(cache_item_id),
            file_directory_path,
            destination_name,
            width,
            height,
        )

    destination_public_path = f'/file/{quote(destination_name)}'

    return destination_public_path


async def _load_image_image(
    file_path: Path,
    media_type: MediaType,
) -> Image:
    # We want to read the image asynchronously and prevent Pillow from keeping too many file
    # descriptors open simultaneously, so we read the image ourselves and store the contents
    # in a synchronous file object.
    async with aiofiles.open(file_path, 'rb') as f:
        image_f = BytesIO(await f.read())
    # Ignore warnings about decompression bombs, because we know where the files come from.
    with warnings.catch_warnings(action='ignore', category=DecompressionBombWarning):
        image = Image.open(image_f, formats=[media_type.subtype])
    return image


async def _load_image_application_pdf(
    file_path: Path,
    media_type: MediaType,
) -> Image:
    # Ignore warnings about decompression bombs, because we know where the files come from.
    with warnings.catch_warnings(action='ignore', category=DecompressionBombWarning):
        image = convert_from_path(file_path, fmt='jpeg')[0]
    return image


@sync
async def _execute_filter_image(
    image_loader: Callable[[Path, MediaType], Awaitable[Image]],
    file_path: Path,
    media_type: MediaType,
    cache_item_file_path: Path,
    destination_directory_path: Path,
    destination_name: str,
    width: int | None,
    height: int | None,
) -> None:
    destination_file_path = destination_directory_path / destination_name
    await makedirs(destination_directory_path, exist_ok=True)
    try:
        await link_or_copy(cache_item_file_path, destination_file_path)
    except FileNotFoundError:
        image = await image_loader(file_path, media_type)
        try:
            if width is not None:
                width = min(width, image.width)
            if height is not None:
                height = min(height, image.height)

            await makedirs(cache_item_file_path.parent, exist_ok=True)
            converted_image = await _execute_filter_image_convert(image, width, height)
            converted_image.save(cache_item_file_path, format=media_type.subtype)
            del converted_image
        finally:
            image.close()
            del image
        await link_or_copy(cache_item_file_path, destination_file_path)


async def _execute_filter_image_convert(
    image: Image,
    width: int | None,
    height: int | None,
) -> Image:
    if width is not None and height is not None:
        return _resizeimage.resize_cover(image, (width, height))
    if width is not None:
        return _resizeimage.resize_width(image, width)
    if height is not None:
        return _resizeimage.resize_height(image, height)
    raise ValueError('Width and height cannot both be None.')


@pass_context
def filter_negotiate_localizeds(context: Context, localizeds: Iterable[Localized]) -> Localized | None:
    """
    Try to find an object whose locale matches the context's current locale.
    """
    from betty.jinja2 import context_localizer

    return negotiate_localizeds(context_localizer(context).locale, list(localizeds))


@pass_context
def filter_sort_localizeds(context: Context, localizeds: Iterable[Localized], localized_attribute: str, sort_attribute: str) -> Iterable[Localized]:
    """
    Sort localized objects.
    """
    from betty.jinja2 import context_localizer

    get_localized_attr = make_attrgetter(
        context.environment, localized_attribute)
    get_sort_attr = make_attrgetter(context.environment, sort_attribute)

    def _get_sort_key(x: Localized) -> Any:
        return get_sort_attr(negotiate_localizeds(context_localizer(context).locale, get_localized_attr(x)))

    return sorted(localizeds, key=_get_sort_key)


@pass_context
def filter_select_localizeds(context: Context, localizeds: Iterable[Localized], include_unspecified: bool = False) -> Iterable[Localized]:
    """
    Select all objects whose locale matches the context's current locale.

    :param include_unspecified: If True, the return value includes all objects that do not have a locale specified.
    """
    from betty.jinja2 import context_localizer

    for localized in localizeds:
        if include_unspecified and localized.locale in {None, 'mis', 'mul', 'und', 'zxx'}:
            yield localized
        if localized.locale is not None and negotiate_locale(context_localizer(context).locale, [localized.locale]) is not None:
            yield localized


@pass_context
def filter_negotiate_dateds(context: Context, dateds: Iterable[Dated], date: Datey | None) -> Dated | None:
    """
    Try to find an object whose date falls in the given date.

    :param date: A date to select by. If None, then today's date is used.
    """
    with suppress(StopIteration):
        return next(filter_select_dateds(context, dateds, date))
    return None


@pass_context
def filter_select_dateds(context: Context, dateds: Iterable[Dated], date: Datey | None) -> Iterator[Dated]:
    """
    Select all objects whose date falls in the given date.

    :param date: A date to select by. If None, then today's date is used.
    """
    if date is None:
        date = context.resolve_or_missing('today')
    return filter(
        lambda dated: dated.date is None or dated.date.comparable and dated.date in date,
        dateds,
    )


def filter_base64(input: str) -> str:
    """
    Base-64-encode a string.
    """
    return b64encode(input.encode('utf-8')).decode('utf-8')


FILTERS = {
    'base64': filter_base64,
    'camel_case_to_kebab_case': camel_case_to_kebab_case,
    'camel_case_to_snake_case': camel_case_to_snake_case,
    'entity_type_name': get_entity_type_name,
    'file': filter_file,
    'flatten': filter_flatten,
    'format_datey': filter_format_datey,
    'format_degrees': filter_format_degrees,
    'image': filter_image,
    'json': filter_json,
    'locale_get_data': get_data,
    'localize': filter_localize,
    'map': filter_map,
    'minimize': minimize,
    'negotiate_dateds': filter_negotiate_dateds,
    'negotiate_localizeds': filter_negotiate_localizeds,
    'none_void': none_void,
    'paragraphs': filter_paragraphs,
    'select_dateds': filter_select_dateds,
    'select_localizeds': filter_select_localizeds,
    'static_url': filter_static_url,
    'sort_localizeds': filter_sort_localizeds,
    'str': str,
    'unique': filter_unique,
    'upper_camel_case_to_lower_camel_case': upper_camel_case_to_lower_camel_case,
    'url': filter_url,
    'void_none': void_none,
    'walk': filter_walk,
}



"""
Provide configuration for the :py:class:`betty.extension.Gramps` extension.
"""
from __future__ import annotations

from pathlib import Path
from typing import Iterable, Any, Self

from betty.config import Configuration, ConfigurationSequence
from betty.serde.dump import minimize, Dump, VoidableDump
from betty.serde.load import Asserter, Fields, RequiredField, Assertions, OptionalField


class FamilyTreeConfiguration(Configuration):
    def __init__(
        self,
        *,
        file_path: Path | None = None,
    ):
        super().__init__()
        self.file_path = file_path

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, FamilyTreeConfiguration):
            return False
        return self._file_path == other.file_path

    @property
    def file_path(self) -> Path | None:
        return self._file_path

    @file_path.setter
    def file_path(self, file_path: Path | None) -> None:
        self._file_path = file_path

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            RequiredField(
                'file',
                Assertions(asserter.assert_path()) | asserter.assert_setattr(configuration, 'file_path'),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return {
            'file': str(self.file_path),
        }


class FamilyTreeConfigurationSequence(ConfigurationSequence[FamilyTreeConfiguration]):
    def update(self, other: Self) -> None:
        self._clear_without_dispatch()
        self.append(*other)

    @classmethod
    def _create_default_item(cls, configuration_key: int) -> FamilyTreeConfiguration:
        return FamilyTreeConfiguration()

    @classmethod
    def _item_type(cls) -> type[FamilyTreeConfiguration]:
        return FamilyTreeConfiguration


class GrampsConfiguration(Configuration):
    def __init__(
        self,
        *,
        family_trees: Iterable[FamilyTreeConfiguration] | None = None,
    ):
        super().__init__()
        self._family_trees = FamilyTreeConfigurationSequence(family_trees)
        self._family_trees.on_change(self)

    @property
    def family_trees(self) -> FamilyTreeConfigurationSequence:
        return self._family_trees

    def update(self, other: Self) -> None:
        self._family_trees.update(other._family_trees)

    @classmethod
    def load(
            cls,
            dump: Dump,
            configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'family_trees',
                Assertions(configuration._family_trees.assert_load(configuration.family_trees)),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'family_trees': self.family_trees.dump(),
        }, True)



"""
Provide project administration for the Graphical User Interface.
"""
from __future__ import annotations

import asyncio
import copy
import re
from asyncio import Task, CancelledError
from contextlib import suppress
from logging import getLogger
from pathlib import Path
from urllib.parse import urlparse

from PyQt6.QtCore import Qt, QThread, QObject
from PyQt6.QtGui import QAction, QCloseEvent
from PyQt6.QtWidgets import QFileDialog, QPushButton, QWidget, QVBoxLayout, QHBoxLayout, QMenu, QStackedLayout, \
    QGridLayout, QCheckBox, QFormLayout, QLabel, QLineEdit, QButtonGroup, QRadioButton, QFrame, QScrollArea, QSizePolicy
from babel import Locale
from babel.localedata import locale_identifiers

from betty import load, generate
from betty.app import App
from betty.app.extension import UserFacingExtension
from betty.asyncio import sync, wait
from betty.gui import get_configuration_file_filter, GuiBuilder, mark_invalid, mark_valid
from betty.gui.app import BettyPrimaryWindow
from betty.gui.error import ExceptionCatcher
from betty.gui.locale import LocalizedObject
from betty.gui.locale import TranslationsLocaleCollector
from betty.gui.logging import LogRecordViewerHandler, LogRecordViewer
from betty.gui.serve import ServeProjectWindow
from betty.gui.text import Text, Caption
from betty.gui.window import BettyMainWindow
from betty.locale import get_display_name, to_locale, Str, Localizable
from betty.model import UserFacingEntity, Entity
from betty.project import LocaleConfiguration, Project
from betty.serde.load import AssertionFailed


class _PaneButton(QPushButton):
    def __init__(self, pane_name: str, project_window: ProjectWindow):
        super().__init__()
        self.setFlat(True)
        self.setProperty('pane-selector', 'true')
        self.setCursor(Qt.CursorShape.PointingHandCursor)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        self._project_window = project_window
        self.released.connect(lambda: self._project_window._navigate_to_pane(pane_name))


class _GenerateHtmlListForm(LocalizedObject, QWidget):
    def __init__(self, app: App):
        super().__init__(app)
        self._form = QFormLayout()
        self.setLayout(self._form)
        self._form_label = QLabel()
        self._form.addRow(self._form_label)
        self._checkboxes_form = QFormLayout()
        self._form.addRow(self._checkboxes_form)
        self._checkboxes: dict[type[UserFacingEntity & Entity], QCheckBox] = {}
        self._update()

    def _update(self) -> None:
        entity_types = list(sorted(
            [
                entity_type
                for entity_type
                in self._app.entity_types
                if issubclass(entity_type, UserFacingEntity)
            ],
            key=lambda x: x.entity_type_label_plural().localize(self._app.localizer),
        ))
        for entity_type in self._checkboxes.keys():
            if entity_type not in entity_types:
                self._form.removeWidget(self._checkboxes[entity_type])
                del self._checkboxes[entity_type]
        for row_i, entity_type in enumerate(entity_types):
            self._update_for_entity_type(entity_type, row_i)

    def _update_for_entity_type(self, entity_type: type[UserFacingEntity & Entity], row_i: int) -> None:
        if entity_type in self._checkboxes:
            self._checkboxes_form.insertRow(row_i, self._checkboxes[entity_type])
            return

        def _update(generate_html_list: bool) -> None:
            self._app.project.configuration.entity_types[entity_type].generate_html_list = generate_html_list
        self._checkboxes[entity_type] = QCheckBox()
        self._checkboxes[entity_type].setChecked(self._app.project.configuration.entity_types[entity_type].generate_html_list)
        self._checkboxes[entity_type].toggled.connect(_update)
        self._update_for_entity_type(entity_type, row_i)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._form_label.setText(self._app.localizer._('Generate entity listing pages'))
        for entity_type in self._app.entity_types:
            if issubclass(entity_type, UserFacingEntity):
                self._checkboxes[entity_type].setText(entity_type.entity_type_label_plural().localize(self._app.localizer))


class _GeneralPane(LocalizedObject, QWidget):
    def __init__(self, app: App):
        super().__init__(app)

        self._form = QFormLayout()
        self.setLayout(self._form)
        self._build_name()
        self._build_title()
        self._build_author()
        self._build_url()
        self._build_lifetime_threshold()
        self._build_mode()
        self._build_clean_urls()
        self._generate_html_list_form = _GenerateHtmlListForm(app)
        self._form.addRow(self._generate_html_list_form)

    def _build_name(self) -> None:
        def _update_configuration_name(name: str) -> None:
            self._app.project.configuration.name = name
        self._configuration_name = QLineEdit()
        self._configuration_name.setText(self._app.project.configuration.name)
        self._configuration_name.textChanged.connect(_update_configuration_name)
        self._configuration_name_label = QLabel()
        self._form.addRow(self._configuration_name_label, self._configuration_name)
        self._configuration_name_caption = Caption()
        self._form.addRow(self._configuration_name_caption)

    def _build_title(self) -> None:
        def _update_configuration_title(title: str) -> None:
            self._app.project.configuration.title = title
        self._configuration_title = QLineEdit()
        self._configuration_title.setText(self._app.project.configuration.title)
        self._configuration_title.textChanged.connect(_update_configuration_title)
        self._configuration_title_label = QLabel()
        self._form.addRow(self._configuration_title_label, self._configuration_title)

    def _build_author(self) -> None:
        def _update_configuration_author(author: str) -> None:
            self._app.project.configuration.author = author
        self._configuration_author = QLineEdit()
        self._configuration_author.setText(str(self._app.project.configuration.author))
        self._configuration_author.textChanged.connect(_update_configuration_author)
        self._configuration_author_label = QLabel()
        self._form.addRow(self._configuration_author_label, self._configuration_author)

    def _build_url(self) -> None:
        def _update_configuration_url(url: str) -> None:
            url_parts = urlparse(url)
            base_url = '%s://%s' % (url_parts.scheme, url_parts.netloc)
            root_path = url_parts.path
            configuration = copy.copy(self._app.project.configuration)
            try:
                configuration.base_url = base_url
                configuration.root_path = root_path
            except AssertionFailed as e:
                mark_invalid(self._configuration_url, str(e))
                return
            self._app.project.configuration.base_url = base_url
            self._app.project.configuration.root_path = root_path
            mark_valid(self._configuration_url)
        self._configuration_url = QLineEdit()
        self._configuration_url.setText(self._app.project.configuration.base_url + self._app.project.configuration.root_path)
        self._configuration_url.textChanged.connect(_update_configuration_url)
        self._configuration_url_label = QLabel()
        self._form.addRow(self._configuration_url_label, self._configuration_url)

    def _build_lifetime_threshold(self) -> None:
        def _update_configuration_lifetime_threshold(lifetime_threshold_value: str) -> None:
            if re.fullmatch(r'^\d+$', lifetime_threshold_value) is None:
                mark_invalid(self._configuration_url, self._app.localizer._('The lifetime threshold must consist of digits only.'))
                return
            lifetime_threshold = int(lifetime_threshold_value)
            try:
                self._app.project.configuration.lifetime_threshold = lifetime_threshold
                mark_valid(self._configuration_lifetime_threshold)
            except AssertionFailed as e:
                mark_invalid(self._configuration_lifetime_threshold, str(e))
        self._configuration_lifetime_threshold = QLineEdit()
        self._configuration_lifetime_threshold.setFixedWidth(32)
        self._configuration_lifetime_threshold.setText(str(self._app.project.configuration.lifetime_threshold))
        self._configuration_lifetime_threshold.textChanged.connect(_update_configuration_lifetime_threshold)
        self._configuration_lifetime_threshold_label = QLabel()
        self._form.addRow(self._configuration_lifetime_threshold_label, self._configuration_lifetime_threshold)
        self._configuration_lifetime_threshold_caption = Caption()
        self._form.addRow(self._configuration_lifetime_threshold_caption)

    def _build_mode(self) -> None:
        def _update_configuration_debug(mode: bool) -> None:
            self._app.project.configuration.debug = mode
        self._development_debug = QCheckBox()
        self._development_debug.setChecked(self._app.project.configuration.debug)
        self._development_debug.toggled.connect(_update_configuration_debug)
        self._form.addRow(self._development_debug)
        self._development_debug_caption = Caption()
        self._form.addRow(self._development_debug_caption)

    def _build_clean_urls(self) -> None:
        def _update_configuration_clean_urls(clean_urls: bool) -> None:
            self._app.project.configuration.clean_urls = clean_urls
        self._clean_urls = QCheckBox()
        self._clean_urls.setChecked(self._app.project.configuration.clean_urls)
        self._clean_urls.toggled.connect(_update_configuration_clean_urls)
        self._form.addRow(self._clean_urls)
        self._clean_urls_caption = Caption()
        self._form.addRow(self._clean_urls_caption)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._configuration_name_label.setText(self._app.localizer._('Name'))
        self._configuration_name_caption.setText(self._app.localizer._("The project's machine name."))
        self._configuration_author_label.setText(self._app.localizer._('Author'))
        self._configuration_url_label.setText(self._app.localizer._('URL'))
        self._configuration_title_label.setText(self._app.localizer._('Title'))
        self._configuration_lifetime_threshold_label.setText(self._app.localizer._('Lifetime threshold'))
        self._configuration_lifetime_threshold_caption.setText(self._app.localizer._('The age at which people are presumed dead.'))
        self._development_debug.setText(self._app.localizer._('Debugging mode'))
        self._development_debug_caption.setText(self._app.localizer._('Output more detailed logs and disable optimizations that make debugging harder.'))
        self._clean_urls.setText(self._app.localizer._('Clean URLs'))
        self._clean_urls_caption.setText(self._app.localizer._('URLs look like <code>/path</code> instead of <code>/path/index.html</code>. This requires a web server that supports it.'))


class _LocalesConfigurationWidget(LocalizedObject, QWidget):
    def __init__(self, app: App):
        super().__init__(app)

        self._layout = QGridLayout()
        self.setLayout(self._layout)
        self._remove_buttons: dict[str, QPushButton | None] = {}
        self._default_buttons: dict[str, QRadioButton] = {}
        self._default_locale_button_group = QButtonGroup()

        self._layout.addWidget(Text('Default locale'))

        locales_data: list[tuple[str, str]] = []
        for locale in self._app.project.configuration.locales:
            locale_name = get_display_name(locale)
            if locale_name is None:
                continue
            locales_data.append((locale, locale_name))
        for locale_index, (locale, locale_name) in enumerate(sorted(
                locales_data,
                key=lambda locale_data: locale_data[1],
        )):
            self._build_locale_configuration(locale, locale_index + 1)

    def _build_locale_configuration(self, locale: str, row_index: int) -> None:
        self._default_buttons[locale] = QRadioButton()
        self._default_buttons[locale].setChecked(locale == self._app.project.configuration.locales.default.locale)

        def _update_locales_configuration_default() -> None:
            self._app.project.configuration.locales.default = locale  # type: ignore[assignment]
        self._default_buttons[locale].clicked.connect(_update_locales_configuration_default)
        self._default_locale_button_group.addButton(self._default_buttons[locale])
        self._layout.addWidget(self._default_buttons[locale], row_index, 0)

        # Allow this locale configuration to be removed only if there are others, and if it is not default one.
        if len(self._app.project.configuration.locales) > 1 and locale != self._app.project.configuration.locales.default.locale:
            def _remove_locale() -> None:
                del self._app.project.configuration.locales[locale]
            remove_button = QPushButton()
            remove_button.released.connect(_remove_locale)
            self._layout.addWidget(remove_button, row_index, 1)
            self._remove_buttons[locale] = remove_button
        else:
            self._remove_buttons[locale] = None

    def _set_translatables(self) -> None:
        super()._set_translatables()
        for locale, button in self._default_buttons.items():
            button.setText(get_display_name(locale, self._app.localizer.locale))
        for button in self._remove_buttons.values():
            if button is not None:
                button.setText(self._app.localizer._('Remove'))


class _LocalizationPane(LocalizedObject, QWidget):
    def __init__(self, app: App):
        super().__init__(app)

        self._layout = QVBoxLayout()
        self.setLayout(self._layout)

        self._add_locale_button = QPushButton()
        self._add_locale_button.released.connect(self._add_locale)
        self._layout.addWidget(self._add_locale_button, 1)

        self._layout.addStretch()

        self._locales_configuration_widget: _LocalesConfigurationWidget
        self._build_locales_configuration()
        self._app.project.configuration.locales.on_change(self._build_locales_configuration)

    def _build_locales_configuration(self) -> None:
        with suppress(AttributeError):
            self._layout.removeWidget(self._locales_configuration_widget)
            self._locales_configuration_widget.close()
            self._locales_configuration_widget.setParent(None)
            del self._locales_configuration_widget
        self._locales_configuration_widget = _LocalesConfigurationWidget(self._app)
        self._layout.insertWidget(0, self._locales_configuration_widget)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._add_locale_button.setText(self._app.localizer._('Add a locale'))

    def _add_locale(self) -> None:
        window = _AddLocaleWindow(self._app, parent=self)
        window.show()


class _AddLocaleWindow(BettyMainWindow):
    window_width = 500
    window_height = 250

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent=parent)

        self._layout = QFormLayout()
        self._widget = QWidget()
        self._widget.setLayout(self._layout)
        self.setCentralWidget(self._widget)

        self._locale_collector = TranslationsLocaleCollector(
            self._app,
            {
                to_locale(Locale.parse(babel_identifier))
                for babel_identifier
                in locale_identifiers()
            },
        )
        for row in self._locale_collector.rows:
            self._layout.addRow(*row)

        self._alias = QLineEdit()
        self._alias_label = QLabel()
        self._layout.addRow(self._alias_label, self._alias)
        self._alias_caption = Caption()
        self._layout.addRow(self._alias_caption)

        buttons_layout = QHBoxLayout()
        self._layout.addRow(buttons_layout)

        self._save_and_close = QPushButton(self._app.localizer._('Save and close'))
        self._save_and_close.released.connect(self._save_and_close_locale)
        buttons_layout.addWidget(self._save_and_close)

        self._cancel = QPushButton(self._app.localizer._('Cancel'))
        self._cancel.released.connect(
            lambda _: self.close()
        )
        buttons_layout.addWidget(self._cancel)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._alias_label.setText(self._app.localizer._('Alias'))
        self._alias_caption.setText(self._app.localizer._('An optional alias is used instead of the locale code to identify this locale, such as in URLs. If US English is the only English language variant on your site, you may want to alias its language code from <code>en-US</code> to <code>en</code>, for instance.'))

    @property
    def window_title(self) -> Localizable:
        return Str._('Add a locale')

    def _save_and_close_locale(self) -> None:
        with ExceptionCatcher(self):
            locale = self._locale_collector.locale.currentData()
            alias: str | None = self._alias.text().strip()
            if alias == '':
                alias = None
            try:
                self._app.project.configuration.locales.append(LocaleConfiguration(
                    locale,
                    alias=alias,
                ))
            except AssertionFailed as e:
                mark_invalid(self._alias, str(e))
                return
            self.close()


class _ExtensionPane(LocalizedObject, QWidget):
    def __init__(self, app: App, extension_type: type[UserFacingExtension]):
        super().__init__(app)
        self._extension_type = extension_type

        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(layout)

        enable_layout = QFormLayout()
        layout.addLayout(enable_layout)

        self._extension_description = Text()
        enable_layout.addRow(self._extension_description)

        def _update_enabled(enabled: bool) -> None:
            with ExceptionCatcher(self):
                if enabled:
                    self._app.project.configuration.extensions.enable(extension_type)
                    extension = self._app.extensions[extension_type]
                    if isinstance(extension, GuiBuilder):
                        layout.addWidget(extension.gui_build())
                else:
                    self._app.project.configuration.extensions.disable(extension_type)
                    extension_gui_item = layout.itemAt(1)
                    if extension_gui_item is not None:
                        extension_gui_widget = extension_gui_item.widget()
                        assert extension_gui_widget is not None
                        layout.removeWidget(extension_gui_widget)
                        extension_gui_widget.close()
                        extension_gui_widget.setParent(None)
                        del extension_gui_widget

        self._extension_enabled = QCheckBox()
        self._extension_enabled_caption = Caption()
        self._set_extension_status()
        self._extension_enabled.toggled.connect(_update_enabled)
        enable_layout.addRow(self._extension_enabled)
        enable_layout.addRow(self._extension_enabled_caption)

        if extension_type in self._app.extensions:
            extension = self._app.extensions[extension_type]
            if isinstance(extension, GuiBuilder):
                layout.addWidget(extension.gui_build())

    def _set_extension_status(self) -> None:
        self._extension_enabled.setDisabled(False)
        self._extension_enabled_caption.setText('')
        if self._extension_type in self._app.extensions:
            self._extension_enabled.setChecked(True)
            disable_requirement = self._app.extensions[self._extension_type].disable_requirement()
            if not disable_requirement.is_met():
                self._extension_enabled.setDisabled(True)
                self._extension_enabled_caption.setText(str(disable_requirement.reduce()))
        else:
            self._extension_enabled.setChecked(False)
            enable_requirement = self._extension_type.enable_requirement()
            if not enable_requirement.is_met():
                self._extension_enabled.setDisabled(True)
                self._extension_enabled_caption.setText(str(enable_requirement.reduce()))

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._extension_description.setText(self._extension_type.description().localize(self._app.localizer))
        self._extension_enabled.setText(self._app.localizer._('Enable {extension}').format(
            extension=self._extension_type.label().localize(self._app.localizer),
        ))


class ProjectWindow(BettyPrimaryWindow):
    def __init__(
        self,
        app: App,
    ):
        super().__init__(app)

        central_widget = QWidget()
        central_layout = QHBoxLayout()
        central_widget.setLayout(central_layout)
        self.setCentralWidget(central_widget)

        self._pane_selectors_container_widget = QWidget()
        self._pane_selectors_container_widget.setFixedWidth(225)

        self._pane_selectors_container = QScrollArea()
        self._pane_selectors_container.setFrameShape(QFrame.Shape.NoFrame)
        self._pane_selectors_container.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._pane_selectors_container.setWidget(self._pane_selectors_container_widget)
        self._pane_selectors_container.setWidgetResizable(True)
        self._pane_selectors_container.setFixedWidth(225)
        central_layout.addWidget(self._pane_selectors_container)

        self._pane_selectors_layout = QVBoxLayout()
        self._pane_selectors_layout.setContentsMargins(0, 0, 25, 0)
        self._pane_selectors_container_widget.setLayout(self._pane_selectors_layout)

        self._builtin_pane_selectors_layout = QVBoxLayout()
        self._pane_selectors_layout.addLayout(self._builtin_pane_selectors_layout)

        pane_selectors_divider = QFrame()
        pane_selectors_divider.setFrameShape(QFrame.Shape.HLine)
        pane_selectors_divider.setFrameShadow(QFrame.Shadow.Sunken)
        self._pane_selectors_layout.addWidget(pane_selectors_divider)

        self._extension_pane_selectors_layout = QVBoxLayout()
        self._pane_selectors_layout.addLayout(self._extension_pane_selectors_layout)

        self._pane_selectors_layout.addStretch()

        self._panes_layout = QStackedLayout()
        central_layout.addLayout(self._panes_layout, 999999999)

        self._panes: dict[str, QWidget] = {}
        self._pane_containers: dict[str, QWidget] = {}
        self._pane_selectors: dict[str, QPushButton] = {}

        self._add_pane('general', _GeneralPane(self._app))
        self._builtin_pane_selectors_layout.addWidget(self._pane_selectors['general'])
        self._navigate_to_pane('general')
        self._add_pane('localization', _LocalizationPane(self._app))
        self._builtin_pane_selectors_layout.addWidget(self._pane_selectors['localization'])
        self._extension_types = [extension_type for extension_type in self._app.discover_extension_types() if issubclass(extension_type, UserFacingExtension)]
        for extension_type in self._extension_types:
            self._add_pane(f'extension-{extension_type.name()}', _ExtensionPane(self._app, extension_type))

        menu_bar = self.menuBar()
        assert menu_bar is not None

        self.project_menu = QMenu()
        menu_bar.addMenu(self.project_menu)
        menu_bar.insertMenu(self.help_menu.menuAction(), self.project_menu)

        self.save_project_as_action = QAction(self)
        self.save_project_as_action.setShortcut('Ctrl+Shift+S')
        self.save_project_as_action.triggered.connect(
            lambda _: self._save_project_as(),
        )
        self.project_menu.addAction(self.save_project_as_action)

        self.generate_action = QAction(self)
        self.generate_action.setShortcut('Ctrl+G')
        self.generate_action.triggered.connect(
            lambda _: self._generate(),
        )
        self.project_menu.addAction(self.generate_action)

        self.serve_action = QAction(self)
        self.serve_action.setShortcut('Ctrl+Alt+S')
        self.serve_action.triggered.connect(
            lambda _: self._serve(),
        )
        self.project_menu.addAction(self.serve_action)

    def _add_pane(self, pane_name: str, pane: QWidget) -> None:
        pane_container = QScrollArea()
        pane_container.setFrameShape(QFrame.Shape.NoFrame)
        pane_container.setWidget(pane)
        pane_container.setWidgetResizable(True)
        pane.setMinimumWidth(300)
        pane.setMaximumWidth(1000)
        self._pane_containers[pane_name] = pane_container
        self._panes[pane_name] = pane
        self._panes_layout.addWidget(pane_container)
        self._pane_selectors[pane_name] = _PaneButton(pane_name, self)

    def _navigate_to_pane(self, pane_name: str) -> None:
        for pane_selector in self._pane_selectors.values():
            pane_selector.setFlat(True)
        self._pane_selectors[pane_name].setFlat(False)
        self._panes_layout.setCurrentWidget(self._pane_containers[pane_name])

    def show(self) -> None:
        self._app.project.configuration.autowrite = True
        super().show()

    def close(self) -> bool:
        self._app.project.configuration.autowrite = False
        return super().close()

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self.project_menu.setTitle('&' + self._app.localizer._('Project'))
        self.save_project_as_action.setText(self._app.localizer._('Save this project as...'))
        self.generate_action.setText(self._app.localizer._('Generate site'))
        self.serve_action.setText(self._app.localizer._('Serve site'))
        self._pane_selectors['general'].setText(self._app.localizer._('General'))
        self._pane_selectors['localization'].setText(self._app.localizer._('Localization'))

        # Sort extension pane selector buttons by their human-readable label.
        extension_pane_selector_labels = [
            (extension_type, extension_type.label().localize(self._app.localizer))
            for extension_type
            in self._extension_types
        ]
        for extension_type, extension_label in sorted(extension_pane_selector_labels, key=lambda x: x[1]):
            extension_pane_name = f'extension-{extension_type.name()}'
            self._pane_selectors[extension_pane_name].setText(extension_type.label().localize(self._app.localizer))
            self._extension_pane_selectors_layout.addWidget(self._pane_selectors[extension_pane_name])

    @property
    def window_title(self) -> Localizable:
        return Str.plain(self._app.project.configuration.title)

    def _save_project_as(self) -> None:
        with ExceptionCatcher(self):
            configuration_file_path_str, __ = QFileDialog.getSaveFileName(
                self,
                self._app.localizer._('Save your project to...'),
                '',
                get_configuration_file_filter().localize(self._app.localizer),
            )
            wait(self._app.project.configuration.write(Path(configuration_file_path_str)))

    def _generate(self) -> None:
        with ExceptionCatcher(self):
            generate_window = _GenerateWindow(self._app, parent=self)
            generate_window.show()

    def _serve(self) -> None:
        with ExceptionCatcher(self):
            serve_window = ServeProjectWindow(self._app, parent=self)
            serve_window.show()


class _GenerateThread(QThread):
    def __init__(self, project: Project, generate_window: _GenerateWindow):
        super().__init__()
        self._project = project
        self._generate_window = generate_window
        self._task: Task[None] | None = None

    @sync
    async def run(self) -> None:
        with suppress(CancelledError):
            self._task = asyncio.create_task(self._generate())
            await self._task

    async def _generate(self) -> None:
        with ExceptionCatcher(self._generate_window, close_parent=True):
            async with App(project=self._project) as app:
                await load.load(app)
                await generate.generate(app)

    def cancel(self) -> None:
        if self._task:
            self._task.cancel()


class _GenerateWindow(BettyMainWindow):
    window_width = 500
    window_height = 100

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent=parent)

        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setWindowFlags(self.windowFlags() ^ Qt.WindowType.WindowCloseButtonHint)

        central_layout = QVBoxLayout()
        central_widget = QWidget()
        central_widget.setLayout(central_layout)
        self.setCentralWidget(central_widget)

        self._log_record_viewer = LogRecordViewer()
        central_layout.addWidget(self._log_record_viewer)

        button_layout = QHBoxLayout()
        central_layout.addLayout(button_layout)

        self._cancel_button = QPushButton()
        self._cancel_button.released.connect(self.close)
        button_layout.addWidget(self._cancel_button)

        self._serve_button = QPushButton()
        self._serve_button.setDisabled(True)
        self._serve_button.released.connect(self._serve)
        button_layout.addWidget(self._serve_button)

        self._log_record_viewer = LogRecordViewer()
        central_layout.addWidget(self._log_record_viewer)

        self._logging_handler = LogRecordViewerHandler(self._log_record_viewer)
        getLogger(__name__).addHandler(self._logging_handler)

        self._thread = _GenerateThread(self._app.project, self)
        self._thread.finished.connect(self._finish_generate)
        self._thread.start()

    @property
    def window_title(self) -> Localizable:
        return Str._('Generating your site...')

    def _serve(self) -> None:
        with ExceptionCatcher(self):
            serve_window = ServeProjectWindow(self._app, parent=self.parent())
            serve_window.show()

    def closeEvent(self, a0: QCloseEvent | None) -> None:
        super().closeEvent(a0)
        self._thread.cancel()
        self._finalize()

    def _finish_generate(self) -> None:
        self._cancel_button.setDisabled(True)
        self._serve_button.setDisabled(False)
        self._finalize()

    def _finalize(self) -> None:
        getLogger(__name__).removeHandler(self._logging_handler)

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._cancel_button.setText(self._app.localizer._('Cancel'))
        self._cancel_button.setText(self._app.localizer._('Cancel'))
        self._serve_button.setText(self._app.localizer._('View site'))



"""
Provide the Documentation API.
"""
import asyncio
import logging
import shutil
from contextlib import suppress, AsyncExitStack
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory

from aiofiles.os import makedirs

from betty import serve
from betty.fs import ROOT_DIRECTORY_PATH
from betty.locale import Str, Localizer
from betty.serve import Server, NoPublicUrlBecauseServerNotStartedError
from betty.subprocess import run_process


async def _build_cache(cache_directory_path: Path) -> Path:
    cache_directory_path /= 'docs'
    if not cache_directory_path.exists():
        await _build(cache_directory_path)
    return cache_directory_path


async def _build(output_directory_path: Path) -> None:
    with suppress(FileExistsError):
        await makedirs(output_directory_path)
    with TemporaryDirectory() as working_directory_path_str:
        working_directory_path = Path(working_directory_path_str)
        # sphinx-apidoc must output to the documentation directory, but because we do not want
        # to 'pollute' that with generated files that must not be committed, do our work in a
        # temporary directory and copy the documentation source files there.
        source_directory_path = working_directory_path / 'source'
        await asyncio.to_thread(shutil.copytree, ROOT_DIRECTORY_PATH / 'documentation', source_directory_path)
        try:

            await run_process(
                [
                    'sphinx-apidoc',
                    '--force',
                    '--separate',
                    '-d',
                    '999',
                    '-o',
                    str(source_directory_path),
                    str(ROOT_DIRECTORY_PATH / 'betty'),
                    str(ROOT_DIRECTORY_PATH / 'betty' / 'tests'),
                ],
                cwd=working_directory_path,
            )
            await run_process(
                [
                    'sphinx-build',
                    '-b',
                    'dirhtml',
                    str(source_directory_path),
                    str(output_directory_path),
                ],
                cwd=working_directory_path,
            )
        except CalledProcessError as e:
            if e.stderr is not None:
                logging.getLogger().error(e.stderr)
            raise


class DocumentationServer(Server):
    def __init__(
        self,
        cache_directory_path: Path,
        *,
        localizer: Localizer,
    ):
        super().__init__(localizer)
        self._cache_directory_path = cache_directory_path
        self._server: Server | None = None
        self._exit_stack = AsyncExitStack()

    @classmethod
    def label(cls) -> Str:
        return Str._('Betty documentation')

    @property
    def public_url(self) -> str:
        if self._server is not None:
            return self._server.public_url
        raise NoPublicUrlBecauseServerNotStartedError()

    async def start(self) -> None:
        await super().start()
        www_directory_path = await _build_cache(self._cache_directory_path)
        self._server = serve.BuiltinServer(www_directory_path, localizer=self._localizer)
        await self._exit_stack.enter_async_context(self._server)
        await self.assert_available()

    async def stop(self) -> None:
        await self._exit_stack.aclose()
        await super().stop()



import sys
from os import path

import betty
from betty import about
from betty.asyncio import wait
from betty.fs import ROOT_DIRECTORY_PATH, FileSystem, ASSETS_DIRECTORY_PATH
from betty.locale import LocalizerRepository

betty_replacements: dict[str, str] = {}

assets = FileSystem()
assets.prepend(ASSETS_DIRECTORY_PATH, 'utf-8')
localizers = LocalizerRepository(assets)
for locale in localizers.locales:
    coverage = wait(localizers.coverage(locale))
    betty_replacements[f'translation-coverage-{locale}'] = str(int(round(100 / (coverage[1] / coverage[0]))))

sys.path.insert(0, path.dirname(path.dirname(betty.__file__)))
project = 'Betty'
version = wait(about.version()) or ''
release = wait(about.version_label())
copyright = 'Bart Feenstra and contributors'
intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'jinja2': ('https://jinja.palletsprojects.com/en/latest/', None),
}
html_favicon = str(ROOT_DIRECTORY_PATH / 'betty' / 'assets' / 'public' / 'static' / 'betty.ico')
html_logo = str(ROOT_DIRECTORY_PATH / 'betty' / 'assets' / 'public' / 'static' / 'betty-32x32.png')
html_context = {
    'display_github': True,
    'github_user': 'bartfeenstra',
    'github_repo': 'betty',
    'github_version': '0.3.x',
    'conf_py_path': '/documentation/',
    'betty_replacements': betty_replacements,
}
html_theme = 'sphinx_immaterial'
highlight_language = 'none'
templates_path = ['_templates']
extensions = [
    'betty.sphinx.extension.replacements',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.viewcode',
    'sphinx_autodoc_typehints',
    'sphinx_design',
    'sphinx_immaterial',
]
nitpicky = True
modindex_common_prefix = ['betty.']

# sphinx.ext.autodoc configuration.
autodoc_member_order = 'alphabetical'

# sphinx_autodoc_typehints configuration.
set_type_checking_flag = True
typehints_fully_qualified = True
always_document_param_types = True



"""
Provide Betty's default theme.
"""
from __future__ import annotations

import asyncio
import logging
import re
from collections import defaultdict
from collections.abc import Sequence, AsyncIterable
from pathlib import Path
from shutil import copy2
from typing import Any, Callable, Iterable, Self, cast

from PyQt6.QtWidgets import QWidget
from aiofiles.os import makedirs
from jinja2 import pass_context
from jinja2.runtime import Context

from betty.app.extension import ConfigurableExtension, Extension, Theme
from betty.config import Configuration
from betty.extension.cotton_candy.search import Index
from betty.extension.npm import _Npm, _NpmBuilder, npm
from betty.functools import walk
from betty.generate import Generator, GenerationContext
from betty.gui import GuiBuilder
from betty.jinja2 import Jinja2Provider, context_app, context_localizer, context_job_context
from betty.locale import Date, Str, Datey
from betty.model import Entity, UserFacingEntity, GeneratedEntityId
from betty.model.ancestry import Event, Person, Presence, is_public, Subject
from betty.model.event_type import StartOfLifeEventType, EndOfLifeEventType
from betty.project import EntityReferenceSequence, EntityReference
from betty.serde.dump import minimize, Dump, VoidableDump
from betty.serde.load import AssertionFailed, Fields, Assertions, OptionalField, Asserter


class _ColorConfiguration(Configuration):
    _HEX_PATTERN = re.compile(r'^#[a-zA-Z0-9]{6}$')

    def __init__(self, hex_value: str):
        super().__init__()
        self._hex: str
        self.hex = hex_value

    def _validate_hex(self, hex_value: str) -> str:
        if not self._HEX_PATTERN.match(hex_value):
            raise AssertionFailed(Str._(
                '"{hex_value}" is not a valid hexadecimal color, such as #ffc0cb.',
                hex_value=hex_value,
            ))
        return hex_value

    @property
    def hex(self) -> str:
        return self._hex

    @hex.setter
    def hex(self, hex_value: str) -> None:
        if not self._HEX_PATTERN.match(hex_value):
            raise AssertionFailed(Str._(
                '"{hex_value}" is not a valid hexadecimal color, such as #ffc0cb.',
                hex_value=hex_value,
            ))
        self._hex = hex_value
        self._dispatch_change()

    def update(self, other: Self) -> None:
        self.hex = other.hex

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        asserter = Asserter()
        hex_value = asserter.assert_str()(dump)
        if configuration is None:
            configuration = cls(hex_value)
        else:
            configuration.hex = hex_value
        return configuration

    def dump(self) -> VoidableDump:
        return self._hex


class CottonCandyConfiguration(Configuration):
    DEFAULT_PRIMARY_INACTIVE_COLOR = '#ffc0cb'
    DEFAULT_PRIMARY_ACTIVE_COLOR = '#ff69b4'
    DEFAULT_LINK_INACTIVE_COLOR = '#149988'
    DEFAULT_LINK_ACTIVE_COLOR = '#2a615a'

    def __init__(
        self,
        *,
        featured_entities: Sequence[EntityReference[UserFacingEntity & Entity]] | None = None,
        primary_inactive_color: str = DEFAULT_PRIMARY_INACTIVE_COLOR,
        primary_active_color: str = DEFAULT_PRIMARY_ACTIVE_COLOR,
        link_inactive_color: str = DEFAULT_LINK_INACTIVE_COLOR,
        link_active_color: str = DEFAULT_LINK_ACTIVE_COLOR,
    ):
        super().__init__()
        self._featured_entities = EntityReferenceSequence['UserFacingEntity & Entity'](featured_entities or ())
        self._featured_entities.on_change(self)
        self._primary_inactive_color = _ColorConfiguration(primary_inactive_color)
        self._primary_inactive_color.on_change(self)
        self._primary_active_color = _ColorConfiguration(primary_active_color)
        self._primary_active_color.on_change(self)
        self._link_inactive_color = _ColorConfiguration(link_inactive_color)
        self._link_inactive_color.on_change(self)
        self._link_active_color = _ColorConfiguration(link_active_color)
        self._link_active_color.on_change(self)

    @property
    def featured_entities(self) -> EntityReferenceSequence[UserFacingEntity & Entity]:
        return self._featured_entities

    @property
    def primary_inactive_color(self) -> _ColorConfiguration:
        return self._primary_inactive_color

    @property
    def primary_active_color(self) -> _ColorConfiguration:
        return self._primary_active_color

    @property
    def link_inactive_color(self) -> _ColorConfiguration:
        return self._link_inactive_color

    @property
    def link_active_color(self) -> _ColorConfiguration:
        return self._link_active_color

    @classmethod
    def load(
        cls,
        dump: Dump,
        configuration: Self | None = None,
    ) -> Self:
        if configuration is None:
            configuration = cls()
        asserter = Asserter()
        asserter.assert_record(Fields(
            OptionalField(
                'featured_entities',
                Assertions(configuration._featured_entities.assert_load(configuration._featured_entities)),
            ),
            OptionalField(
                'primary_inactive_color',
                Assertions(configuration._primary_inactive_color.assert_load(configuration._primary_inactive_color)),
            ),
            OptionalField(
                'primary_active_color',
                Assertions(configuration._primary_active_color.assert_load(configuration._primary_active_color)),
            ),
            OptionalField(
                'link_inactive_color',
                Assertions(configuration._link_inactive_color.assert_load(configuration._link_inactive_color)),
            ),
            OptionalField(
                'link_active_color',
                Assertions(configuration._link_active_color.assert_load(configuration._link_active_color)),
            ),
        ))(dump)
        return configuration

    def dump(self) -> VoidableDump:
        return minimize({
            'featured_entities': self.featured_entities.dump(),
            'primary_inactive_color': self._primary_inactive_color.dump(),
            'primary_active_color': self._primary_active_color.dump(),
            'link_inactive_color': self._link_inactive_color.dump(),
            'link_active_color': self._link_active_color.dump(),
        })


class _CottonCandy(Theme, ConfigurableExtension[CottonCandyConfiguration], Generator, GuiBuilder, _NpmBuilder, Jinja2Provider):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @classmethod
    def label(cls) -> Str:
        return Str.plain('Cotton Candy')

    @classmethod
    def default_configuration(cls) -> CottonCandyConfiguration:
        return CottonCandyConfiguration()

    @classmethod
    def description(cls) -> Str:
        return Str._("Cotton Candy is Betty's default theme.")

    def gui_build(self) -> QWidget:
        from betty.extension.cotton_candy.gui import _CottonCandyGuiWidget

        return _CottonCandyGuiWidget(self._app)

    @property
    def globals(self) -> dict[str, Any]:
        return {
            'search_index': _global_search_index,
        }

    @property
    def filters(self) -> dict[str, Callable[..., Any]]:
        return {
            'person_timeline_events': lambda person: person_timeline_events(
                person,
                self.app.project.configuration.lifetime_threshold
            ),
            'person_descendant_families': person_descendant_families,
        }

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await npm(('run', 'webpack'), cwd=working_directory_path)
        await self._copy_npm_build(working_directory_path / 'webpack-build', assets_directory_path)
        logging.getLogger(__name__).info(self._app.localizer._('Built the Cotton Candy front-end assets.'))

    async def _copy_npm_build(self, source_directory_path: Path, destination_directory_path: Path) -> None:
        await makedirs(destination_directory_path, exist_ok=True)
        await asyncio.to_thread(copy2, source_directory_path / 'cotton_candy.css', destination_directory_path / 'cotton_candy.css')
        await asyncio.to_thread(copy2, source_directory_path / 'cotton_candy.js', destination_directory_path / 'cotton_candy.js')

    async def generate(self, job_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await makedirs(self.app.project.configuration.www_directory_path, exist_ok=True)
        await self._copy_npm_build(assets_directory_path, self.app.project.configuration.www_directory_path)


@pass_context
async def _global_search_index(context: Context) -> AsyncIterable[dict[str, str]]:
    return Index(
        context_app(context),
        context_job_context(context),
        context_localizer(context),
    ).build()


def _is_person_timeline_presence(presence: Presence) -> bool:
    if presence.private:
        return False
    if not presence.event:
        return False
    if not presence.event.date:
        return False
    if not presence.event.date.comparable:
        return False
    return True


def person_timeline_events(person: Person, lifetime_threshold: int) -> Iterable[Event]:
    """
    Gather all events for a person's timeline.
    """
    seen = []
    for event in _person_timeline_events(person, lifetime_threshold):
        if event in seen:
            continue
        seen.append(event)
        yield event


def person_descendant_families(person: Person) -> Iterable[tuple[Sequence[Person], Sequence[Person]]]:
    """
    Gather a person's families they are a parent in.
    """
    parents = {}
    children = defaultdict(set)
    for child in person.children:
        family = tuple(sorted(map(lambda parent: parent.id, child.parents)))
        if family not in parents:
            parents[family] = tuple(child.parents)
        children[family].add(child)
    yield from zip(parents.values(), children.values())


def _person_timeline_events(person: Person, lifetime_threshold: int) -> Iterable[Event]:
    # Collect all associated events for a person.
    # Start with the person's own events for which their presence is public.
    for presence in person.presences:
        if _is_person_timeline_presence(presence):
            assert presence.event is not None
            yield presence.event
        continue

    # If the person has start- or end-of-life events, we use those to constrain associated people's events.
    start_dates = []
    end_dates = []
    for presence in person.presences:
        if not _is_person_timeline_presence(presence):
            continue
        assert presence.event is not None
        assert presence.event.date is not None
        if not isinstance(presence.role, Subject):
            continue
        if issubclass(presence.event.event_type, StartOfLifeEventType):
            start_dates.append(presence.event.date)
        if issubclass(presence.event.event_type, EndOfLifeEventType):
            end_dates.append(presence.event.date)
    start_date = sorted(start_dates)[0] if start_dates else None
    end_date = sorted(end_dates)[0] if end_dates else None

    # If an end-of-life event exists, but no start-of-life event, create a start-of-life date based on the end date,
    # minus the lifetime threshold.
    if start_date is None and end_date is not None:
        if isinstance(end_date, Date):
            start_date_reference = end_date
        else:
            if end_date.end is not None and end_date.end.comparable:
                start_date_reference = end_date.end
            else:
                assert end_date.start is not None
                start_date_reference = end_date.start
        assert start_date_reference.year is not None
        start_date = Date(
            start_date_reference.year - lifetime_threshold,
            start_date_reference.month,
            start_date_reference.day,
            start_date_reference.fuzzy,
        )

    # If a start-of-life event exists, but no end-of-life event, create an end-of-life date based on the start date,
    # plus the lifetime threshold.
    if end_date is None and start_date is not None:
        if isinstance(start_date, Date):
            end_date_reference = start_date
        else:
            if start_date.start and start_date.start.comparable:
                end_date_reference = start_date.start
            else:
                assert start_date.end is not None
                end_date_reference = start_date.end
        assert end_date_reference.year is not None
        end_date = Date(
            end_date_reference.year + lifetime_threshold,
            end_date_reference.month,
            end_date_reference.day,
            end_date_reference.fuzzy,
        )

    if start_date is None or end_date is None:
        reference_dates = list(sorted(
            cast(Datey, cast(Event, presence.event).date)
            for presence
            in person.presences
            if _is_person_timeline_presence(presence)
        ))
        if reference_dates:
            if not start_date:
                start_date = reference_dates[0]
            if not end_date:
                end_date = reference_dates[-1]

    if start_date is not None and end_date is not None:
        associated_people = filter(is_public, (
            # All ancestors.
            *walk(person, 'parents'),
            # All descendants.
            *walk(person, 'children'),
            # All siblings.
            *person.siblings,
        ))
        for associated_person in associated_people:
            # For associated events, we are only interested in people's start- or end-of-life events.
            for associated_presence in associated_person.presences:
                if not associated_presence.event or not issubclass(associated_presence.event.event_type, (StartOfLifeEventType, EndOfLifeEventType)):
                    continue
                if isinstance(associated_presence.event.id, GeneratedEntityId):
                    continue
                if not isinstance(associated_presence.role, Subject):
                    continue
                if not _is_person_timeline_presence(associated_presence):
                    continue
                if not associated_presence.event.date:
                    continue
                if associated_presence.event.date < start_date:
                    continue
                if associated_presence.event.date > end_date:
                    continue
                yield associated_presence.event



"""Integrate Betty with `ReDoc <https://redocly.com/redoc/>`_."""
from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from shutil import copy2

from aiofiles.os import makedirs

from betty.app.extension import Extension, UserFacingExtension
from betty.extension.npm import _Npm, _NpmBuilder, _NpmBuilderCacheScope
from betty.generate import Generator, GenerationContext
from betty.locale import Str


class _HttpApiDoc(UserFacingExtension, Generator, _NpmBuilder):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await asyncio.to_thread(copy2, working_directory_path / 'node_modules' / 'redoc' / 'bundles' / 'redoc.standalone.js', assets_directory_path / 'http-api-doc.js')
        logging.getLogger(__name__).info(self._app.localizer._('Built the HTTP API documentation.'))

    @classmethod
    def npm_cache_scope(cls) -> _NpmBuilderCacheScope:
        return _NpmBuilderCacheScope.BETTY

    async def generate(self, job_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await makedirs(self.app.project.configuration.www_directory_path, exist_ok=True)
        await asyncio.to_thread(copy2, assets_directory_path / 'http-api-doc.js', self.app.project.configuration.www_directory_path / 'http-api-doc.js')

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @classmethod
    def label(cls) -> Str:
        return Str._('HTTP API Documentation')

    @classmethod
    def description(cls) -> Str:
        return Str._('Display the HTTP API documentation in a user-friendly way using <a href="https://github.com/Redocly/redoc">ReDoc</a>.')



"""
Build nginx and Docker artifacts, such as configuration files.
"""
import asyncio
from pathlib import Path
from shutil import copyfile
from urllib.parse import urlparse

import aiofiles
from aiofiles.os import makedirs
from jinja2 import FileSystemLoader

from betty.app import App
from betty.path import rootname


async def generate_configuration_file(
    app: App,
    destination_file_path: Path | None = None,
    www_directory_path: str | None = None,
    https: bool | None = None,
) -> None:
    """
    Generate an ``nginx.conf`` file to the given destination path.
    """
    from betty.extension import Nginx

    data = {
        'server_name': urlparse(app.project.configuration.base_url).netloc,
        'www_directory_path': www_directory_path or app.extensions[Nginx].www_directory_path,
        'https': https or app.extensions[Nginx].https,
    }
    if destination_file_path is None:
        destination_file_path = app.project.configuration.output_directory_path / 'nginx' / 'nginx.conf'
    root_path = rootname(Path(__file__))
    configuration_file_template_name = '/'.join((Path(__file__).parent / 'assets' / 'nginx.conf.j2').relative_to(root_path).parts)
    template = FileSystemLoader(root_path).load(app.jinja2_environment, configuration_file_template_name, app.jinja2_environment.globals)
    await makedirs(destination_file_path.parent, exist_ok=True)
    configuration_file_contents = await template.render_async(data)
    async with aiofiles.open(destination_file_path, 'w', encoding='utf-8') as f:
        await f.write(configuration_file_contents)


async def generate_dockerfile_file(app: App, destination_file_path: Path | None = None) -> None:
    """
    Generate a ``Dockerfile`` to the given destination path.
    """
    if destination_file_path is None:
        destination_file_path = app.project.configuration.output_directory_path / 'nginx' / 'Dockerfile'
    await makedirs(destination_file_path.parent, exist_ok=True)
    await asyncio.to_thread(copyfile, Path(__file__).parent / 'assets' / 'Dockerfile', destination_file_path)
    await asyncio.to_thread(copyfile, Path(__file__).parent / 'assets' / 'content_negotiation.lua', destination_file_path.parent / 'content_negotiation.lua')



"""Provide interactive family trees by integrating Betty with `Cytoscape.js <https://cytoscape.org/>`_."""
from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from shutil import copy2

from aiofiles.os import makedirs

from betty.app.extension import Extension, UserFacingExtension
from betty.extension.npm import _Npm, _NpmBuilder, npm, _NpmBuilderCacheScope
from betty.generate import Generator, GenerationContext
from betty.html import CssProvider, JsProvider
from betty.locale import Str


class _Trees(UserFacingExtension, CssProvider, JsProvider, Generator, _NpmBuilder):
    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return {_Npm}

    async def npm_build(self, working_directory_path: Path, assets_directory_path: Path) -> None:
        await self.app.extensions[_Npm].install(type(self), working_directory_path)
        await npm(('run', 'webpack'), cwd=working_directory_path)
        await self._copy_npm_build(working_directory_path / 'webpack-build', assets_directory_path)
        logging.getLogger(__name__).info(self._app.localizer._('Built the interactive family trees.'))

    async def _copy_npm_build(self, source_directory_path: Path, destination_directory_path: Path) -> None:
        await makedirs(destination_directory_path, exist_ok=True)
        await asyncio.to_thread(copy2, source_directory_path / 'trees.css', destination_directory_path / 'trees.css')
        await asyncio.to_thread(copy2, source_directory_path / 'trees.js', destination_directory_path / 'trees.js')

    @classmethod
    def npm_cache_scope(cls) -> _NpmBuilderCacheScope:
        return _NpmBuilderCacheScope.BETTY

    async def generate(self, job_context: GenerationContext) -> None:
        assets_directory_path = await self.app.extensions[_Npm].ensure_assets(self)
        await self._copy_npm_build(assets_directory_path, self.app.project.configuration.www_directory_path)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        return Path(__file__).parent / 'assets'

    @property
    def public_css_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('trees.css'),
        ]

    @property
    def public_js_paths(self) -> list[str]:
        return [
            self.app.static_url_generator.generate('trees.js'),
        ]

    @classmethod
    def label(cls) -> Str:
        return Str._('Trees')

    @classmethod
    def description(cls) -> Str:
        return Str._('Display interactive family trees using <a href="https://cytoscape.org/">Cytoscape</a>.')



"""
Integrate Betty's Graphical User Interface with the Serve API.
"""
from __future__ import annotations

from typing import Any

from PyQt6.QtCore import Qt, QThread, pyqtSignal, QObject
from PyQt6.QtWidgets import QVBoxLayout, QWidget, QPushButton

from betty import documentation, fs
from betty.app import App
from betty.asyncio import sync
from betty.extension import demo
from betty.gui.error import ExceptionCatcher
from betty.gui.text import Text
from betty.gui.window import BettyMainWindow
from betty.locale import Str, Localizable
from betty.project import Project
from betty.serve import Server, AppServer


class _ServeThread(QThread):
    server_started = pyqtSignal()

    def __init__(self, project: Project, server: Server, serve_window: _ServeWindow, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._project = project
        self._server = server
        self._serve_window = serve_window
        self._app: App | None = None

    @property
    def server(self) -> Server:
        return self._server

    @sync
    async def run(self) -> None:
        with ExceptionCatcher(self._serve_window, close_parent=True):
            async with App(project=self._project) as self._app:
                await self._server.start()
                self.server_started.emit()
                await self._server.show()

    @sync
    async def stop(self) -> None:
        await self._server.stop()


class _ServeWindow(BettyMainWindow):
    window_width = 500
    window_height = 100

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent=parent)

        self.__thread: _ServeThread | None = None

        self._central_layout = QVBoxLayout()
        central_widget = QWidget()
        central_widget.setLayout(self._central_layout)
        self.setCentralWidget(central_widget)

        self._loading_instruction = Text(self._app.localizer._('Loading...'))
        self._loading_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(self._loading_instruction)

    def _server(self) -> Server:
        raise NotImplementedError(repr(self))

    @property
    def _thread(self) -> _ServeThread:
        if self.__thread is None:
            raise RuntimeError('This window has not been shown yet.')
        return self.__thread

    def _build_instruction(self) -> str:
        raise NotImplementedError(repr(self))

    def _server_started(self) -> None:
        # The server may have been stopped before this method was called.
        if self.__thread is None:
            return

        self._loading_instruction.close()

        instance_instruction = Text(self._build_instruction())
        instance_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(instance_instruction)

        general_instruction = Text(self._app.localizer._('Keep this window open to keep the site running.'))
        general_instruction.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._central_layout.addWidget(general_instruction)

        stop_server_button = QPushButton(self._app.localizer._('Stop the site'), self)
        stop_server_button.released.connect(
            self.close,
        )
        self._central_layout.addWidget(stop_server_button)

    def show(self) -> None:
        super().show()
        # Explicitly activate this window in case it existed and was shown before, but requested again.
        self.activateWindow()
        self._start()

    def _start(self) -> None:
        if self.__thread is None:
            self.__thread = _ServeThread(self._app.project, self._server(), self)
            self.__thread.server_started.connect(self._server_started)
            self.__thread.start()

    def close(self) -> bool:
        self._stop()
        return super().close()

    def _stop(self) -> None:
        if self.__thread is not None:
            self.__thread.stop()
        self.__thread = None


class ServeProjectWindow(_ServeWindow):
    def _server(self) -> Server:
        return AppServer.get(self._app)

    @property
    def window_title(self) -> Localizable:
        return Str._('Serving your site...')

    def _build_instruction(self) -> str:
        return self._app.localizer._('You can now view your site at <a href="{url}">{url}</a>.').format(
            url=self._thread.server.public_url,
        )


class ServeDemoWindow(_ServeWindow):
    def _server(self) -> Server:
        return demo.DemoServer()

    def _build_instruction(self) -> str:
        return self._app.localizer._('You can now view a Betty demonstration site at <a href="{url}">{url}</a>.').format(
            url=self._thread.server.public_url,
        )

    @property
    def window_title(self) -> Localizable:
        return Str._('Serving the Betty demo...')


class ServeDocsWindow(_ServeWindow):
    def _server(self) -> Server:
        return documentation.DocumentationServer(
            fs.CACHE_DIRECTORY_PATH,
            localizer=self._app.localizer,
        )

    def _build_instruction(self) -> str:
        return self._app.localizer._('You can now view the documentation at <a href="{url}">{url}</a>.').format(
            url=self._thread.server.public_url,
        )

    @property
    def window_title(self) -> Localizable:
        return Str._('Serving the Betty documentation...')



"""Provide the built-in extensions."""

from __future__ import annotations

from betty.extension.cotton_candy import _CottonCandy
from betty.extension.demo import _Demo
from betty.extension.deriver import _Deriver
from betty.extension.gramps import _Gramps
from betty.extension.http_api_doc import _HttpApiDoc
from betty.extension.maps import _Maps
from betty.extension.nginx import _Nginx
from betty.extension.privatizer import _Privatizer
from betty.extension.trees import _Trees
from betty.extension.wikipedia import _Wikipedia


class CottonCandy(_CottonCandy):
    pass


class Demo(_Demo):
    pass


class Deriver(_Deriver):
    pass


class Gramps(_Gramps):
    pass


class HttpApiDoc(_HttpApiDoc):
    pass


class Maps(_Maps):
    pass


class Nginx(_Nginx):
    pass


class Privatizer(_Privatizer):
    pass


class Trees(_Trees):
    pass


class Wikipedia(_Wikipedia):
    pass



"""
Provide Cotton Candy's search functionality.
"""
from __future__ import annotations

from collections.abc import AsyncIterable
from typing import Any

from betty.app import App
from betty.locale import Localizer
from betty.model import get_entity_type_name, Entity
from betty.model.ancestry import Person, Place, File
from betty.string import camel_case_to_snake_case
from betty.job import Context


class Index:
    def __init__(
        self,
        app: App,
        job_context: Context | None,
        localizer: Localizer,
    ):
        self._app = app
        self._job_context = job_context
        self._localizer = localizer

    async def build(self) -> AsyncIterable[dict[str, str]]:
        async for entry in self._build_people():
            yield entry
        async for entry in self._build_places():
            yield entry
        async for entry in self._build_files():
            yield entry

    async def _build_people(self) -> AsyncIterable[dict[str, str]]:
        for person in self._app.project.ancestry[Person]:
            entry = await self._build_person(person)
            if entry is not None:
                yield entry

    async def _build_places(self) -> AsyncIterable[dict[str, str]]:
        for place in self._app.project.ancestry[Place]:
            entry = await self._build_place(place)
            if entry is not None:
                yield entry

    async def _build_files(self) -> AsyncIterable[dict[str, str]]:
        for file in self._app.project.ancestry[File]:
            entry = await self._build_file(file)
            if entry is not None:
                yield entry

    async def _render_entity(self, entity: Entity) -> str:
        entity_type_name = get_entity_type_name(entity)
        return await self._app.jinja2_environment.select_template([
            f'search/result-{camel_case_to_snake_case(entity_type_name)}.html.j2',
            'search/result.html.j2',
        ]).render_async({
            'job_context': self._job_context,
            'localizer': self._localizer,
            'entity': entity,
        })

    async def _build_person(self, person: Person) -> dict[Any, Any] | None:
        if person.private:
            return None

        names = []
        for name in person.names:
            if name.individual is not None:
                names.append(name.individual.lower())
            if name.affiliation is not None:
                names.append(name.affiliation.lower())
        if not names:
            return None
        return {
            'text': ' '.join(names),
            'result': await self._render_entity(person),
        }

    async def _build_place(self, place: Place) -> dict[Any, Any] | None:
        if place.private:
            return None

        return {
            'text': ' '.join(map(lambda x: x.name.lower(), place.names)),
            'result': await self._render_entity(place),
        }

    async def _build_file(self, file: File) -> dict[Any, Any] | None:
        if file.private:
            return None

        if not file.description:
            return None
        return {
            'text': file.description.lower(),
            'result': await self._render_entity(file),
        }



"""Provide Betty's extension API."""
from __future__ import annotations

import functools
from collections import defaultdict
from importlib.metadata import entry_points, EntryPoint
from pathlib import Path
from typing import Any, TypeVar, Iterable, TYPE_CHECKING, Generic, \
    Iterator, Sequence, Self

from betty.app.extension.requirement import Requirement, AllRequirements
from betty.asyncio import gather
from betty.config import ConfigurationT, Configurable
from betty.dispatch import Dispatcher, TargetedDispatcher
from betty.importlib import import_any
from betty.locale import Str

if TYPE_CHECKING:
    from betty.app import App


class ExtensionError(BaseException):
    pass


class ExtensionTypeError(ExtensionError, ValueError):
    pass


class ExtensionTypeImportError(ExtensionTypeError, ImportError):
    """
    Raised when an alleged extension type cannot be imported.
    """

    def __init__(self, extension_type_name: str):
        super().__init__(f'Cannot find and import an extension with name "{extension_type_name}".')


class ExtensionTypeInvalidError(ExtensionTypeError, ImportError):
    """
    Raised for types that are not valid extension types.
    """

    def __init__(self, extension_type: type):
        super().__init__(f'{extension_type.__module__}.{extension_type.__name__} is not an extension type class. Extension types must extend {Extension.__module__}.{Extension.__name__}.')


class CyclicDependencyError(ExtensionError, RuntimeError):
    def __init__(self, extension_types: Iterable[type[Extension]]):
        extension_names = ', '.join([extension.name() for extension in extension_types])
        super().__init__(f'The following extensions have cyclic dependencies: {extension_names}')


class Dependencies(AllRequirements):
    def __init__(self, dependent_type: type[Extension]):
        dependency_requirements = []
        for dependency_type in dependent_type.depends_on():
            try:
                dependency_requirement = dependency_type.enable_requirement()
            except RecursionError:
                raise CyclicDependencyError([dependency_type])
            else:
                dependency_requirements.append(dependency_requirement)
        super().__init__(*dependency_requirements)
        self._dependent_type = dependent_type

    @classmethod
    def for_dependent(cls, dependent_type: type[Extension]) -> Self:
        return cls(dependent_type)

    def summary(self) -> Str:
        return Str._(
            '{dependent_label} requires {dependency_labels}.',
            dependent_label=format_extension_type(self._dependent_type),
            dependency_labels=Str.call(
                lambda localizer: ', '.join(
                    map(
                        lambda extension_type: format_extension_type(extension_type).localize(localizer),
                        self._dependent_type.depends_on(),
                    ),
                ),
            ),
        )


class Dependents(Requirement):
    def __init__(self, dependency: Extension, dependents: Sequence[Extension]):
        super().__init__()
        self._dependency = dependency
        self._dependents = dependents

    def summary(self) -> Str:
        return Str._(
            '{dependency_label} is required by {dependency_labels}.',
            dependency_label=format_extension_type(type(self._dependency)),
            dependent_labels=Str.call(lambda localizer: ', '.join([
                format_extension_type(type(dependent)).localize(localizer)
                for dependent
                in self._dependents
            ])),
        )

    def is_met(self) -> bool:
        # This class is never instantiated unless there is at least one enabled dependent, which means this requirement
        # is always met.
        return True

    @classmethod
    def for_dependency(cls, dependency: Extension) -> Self:
        dependents = [
            dependency.app.extensions[extension_type]
            for extension_type
            in discover_extension_types()
            if dependency.__class__ in extension_type.depends_on() and extension_type in dependency.app.extensions
        ]
        return cls(dependency, dependents)


class Extension:
    """
    Integrate optional functionality with the Betty app.
    """

    def __init__(self, app: App, *args: Any, **kwargs: Any):
        assert type(self) is not Extension
        super().__init__(*args, **kwargs)
        self._app = app

    @classmethod
    def name(cls) -> str:
        return '%s.%s' % (cls.__module__, cls.__name__)

    @classmethod
    def depends_on(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def comes_after(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def comes_before(cls) -> set[type[Extension]]:
        return set()

    @classmethod
    def enable_requirement(cls) -> Requirement:
        """
        Define the requirement for this extension to be enabled.

        This defaults to the extension's dependencies.
        """
        return Dependencies.for_dependent(cls)

    def disable_requirement(self) -> Requirement:
        """
        Define the requirement for this extension to be disabled.

        This defaults to the extension's dependents.
        """
        return Dependents.for_dependency(self)

    @classmethod
    def assets_directory_path(cls) -> Path | None:
        """
        Return the path on disk where the extension's assets are located.

        This may be anywhere in your Python package.
        """
        return None

    @property
    def app(self) -> App:
        return self._app


ExtensionT = TypeVar('ExtensionT', bound=Extension)


class UserFacingExtension(Extension):
    @classmethod
    def label(cls) -> Str:
        raise NotImplementedError(repr(cls))

    @classmethod
    def description(cls) -> Str:
        raise NotImplementedError(repr(cls))


class Theme(UserFacingExtension):
    pass


@functools.singledispatch
def get_extension_type(extension_type_definition: str | type[Extension] | Extension) -> type[Extension]:
    """
    Get the extension type for an extension, extension type, or extension type name.
    """
    raise ExtensionTypeError(f'Cannot get the extension type for "{extension_type_definition}".')


@get_extension_type.register(str)
def get_extension_type_by_name(extension_type_name: str) -> type[Extension]:
    """
    Get the extension type for an extension type name.
    """
    try:
        extension_type = import_any(extension_type_name)
    except ImportError:
        raise ExtensionTypeImportError(extension_type_name) from None
    return get_extension_type(extension_type)


@get_extension_type.register(type)
def get_extension_type_by_type(extension_type: type) -> type[Extension]:
    """
    Get the extension type for an extension type.
    """
    if issubclass(extension_type, Extension):
        return extension_type
    raise ExtensionTypeInvalidError(extension_type)


@get_extension_type.register(Extension)
def get_extension_type_by_extension(extension: Extension) -> type[Extension]:
    """
    Get the extension type for an extension.
    """
    return get_extension_type(type(extension))


def format_extension_type(extension_type: type[Extension]) -> Str:
    """
    Format an extension type to a human-readable label.
    """
    if issubclass(extension_type, UserFacingExtension):
        return Str.call(lambda localizer: f'{extension_type.label().localize(localizer)} ({extension_type.name()})')
    return Str.plain(extension_type.name())


class ConfigurableExtension(Extension, Generic[ConfigurationT], Configurable[ConfigurationT]):
    def __init__(self, *args: Any, configuration: ConfigurationT | None = None, **kwargs: Any):
        assert type(self) is not ConfigurableExtension
        super().__init__(*args, **kwargs)
        self._configuration = configuration or self.default_configuration()

    @classmethod
    def default_configuration(cls) -> ConfigurationT:
        raise NotImplementedError(repr(cls))


class Extensions:
    def __getitem__(self, extension_type: type[ExtensionT] | str) -> ExtensionT:
        raise NotImplementedError(repr(self))

    def __iter__(self) -> Iterator[Iterator[Extension]]:
        raise NotImplementedError(repr(self))

    def flatten(self) -> Iterator[Extension]:
        raise NotImplementedError(repr(self))

    def __contains__(self, extension_type: type[Extension] | str | Any) -> bool:
        raise NotImplementedError(repr(self))


class ListExtensions(Extensions):
    def __init__(self, extensions: list[list[Extension]]):
        super().__init__()
        self._extensions = extensions

    def __getitem__(self, extension_type: type[ExtensionT] | str) -> ExtensionT:
        if isinstance(extension_type, str):
            extension_type = import_any(extension_type)
        for extension in self.flatten():
            if type(extension) is extension_type:
                return extension  # type: ignore[return-value]
        raise KeyError(f'Unknown extension of type "{extension_type}"')

    def __iter__(self) -> Iterator[Iterator[Extension]]:
        # Use a generator so we discourage calling code from storing the result.
        for batch in self._extensions:
            yield (extension for extension in batch)

    def flatten(self) -> Iterator[Extension]:
        for batch in self:
            yield from batch

    def __contains__(self, extension_type: type[Extension] | str) -> bool:
        if isinstance(extension_type, str):
            try:
                extension_type = import_any(extension_type)
            except ImportError:
                return False
        for extension in self.flatten():
            if type(extension) is extension_type:
                return True
        return False


class ExtensionDispatcher(Dispatcher):
    def __init__(self, extensions: Extensions):
        self._extensions = extensions

    def dispatch(self, target_type: type[Any]) -> TargetedDispatcher:
        target_method_names = [method_name for method_name in dir(target_type) if not method_name.startswith('_')]
        if len(target_method_names) != 1:
            raise ValueError(f"A dispatch's target type must have a single method to dispatch to, but {target_type} has {len(target_method_names)}.")
        target_method_name = target_method_names[0]

        async def _dispatch(*args: Any, **kwargs: Any) -> list[Any]:
            return [
                result
                for target_extension_batch
                in self._extensions
                for result
                in await gather(*(
                    getattr(target_extension, target_method_name)(*args, **kwargs)
                    for target_extension in target_extension_batch
                    if isinstance(target_extension, target_type)
                ))
            ]
        return _dispatch


ExtensionTypeGraph = dict[type[Extension], set[type[Extension]]]


def build_extension_type_graph(extension_types: Iterable[type[Extension]]) -> ExtensionTypeGraph:
    """
    Build a dependency graph of the given extension types.
    """
    extension_types_graph: ExtensionTypeGraph = defaultdict(set)
    # Add dependencies to the extension graph.
    for extension_type in extension_types:
        _extend_extension_type_graph(extension_types_graph, extension_type)
    # Now all dependencies have been collected, extend the graph with optional extension orders.
    for extension_type in extension_types:
        for before in extension_type.comes_before():
            if before in extension_types_graph:
                extension_types_graph[before].add(extension_type)
        for after in extension_type.comes_after():
            if after in extension_types_graph:
                extension_types_graph[extension_type].add(after)

    return extension_types_graph


def _extend_extension_type_graph(graph: ExtensionTypeGraph, extension_type: type[Extension]) -> None:
    dependencies = extension_type.depends_on()
    # Ensure each extension type appears in the graph, even if they're isolated.
    graph.setdefault(extension_type, set())
    for dependency in dependencies:
        seen_dependency = dependency in graph
        graph[extension_type].add(dependency)
        if not seen_dependency:
            _extend_extension_type_graph(graph, dependency)


def discover_extension_types() -> set[type[Extension]]:
    """
    Gather the available extension types.
    """
    betty_entry_points: Sequence[EntryPoint]
    betty_entry_points = entry_points(  # type: ignore[assignment, unused-ignore]
        group='betty.extensions',  # type: ignore[call-arg, unused-ignore]
    )
    return {import_any(betty_entry_point.value) for betty_entry_point in betty_entry_points}



"""
Provide `media type <https://en.wikipedia.org/wiki/Media_type>`_ handling utilities.
"""
from __future__ import annotations

from email.message import EmailMessage
from typing import Any


class InvalidMediaType(ValueError):
    pass


class MediaType:
    def __init__(self, media_type: str):
        self._str = media_type
        message = EmailMessage()
        message['Content-Type'] = media_type
        type_part = message.get_content_type()
        # EmailMessage.get_content_type() always returns a type, and will fall back to alternatives if the header is
        # invalid.
        if not media_type.startswith(type_part):
            raise InvalidMediaType(f'"{media_type}" is not a valid media type.')
        self._parameters: dict[str, str] = dict(message['Content-Type'].params)
        self._type, self._subtype = type_part.split('/')
        if not self._subtype:
            raise InvalidMediaType('The subtype must not be empty.')

    @property
    def type(self) -> str:
        return self._type

    @property
    def subtype(self) -> str:
        return self._subtype

    @property
    def subtypes(self) -> list[str]:
        return self._subtype.split('+')[0].split('.')

    @property
    def suffix(self) -> str | None:
        if '+' not in self._subtype:
            return None

        return self._subtype.split('+')[-1]

    @property
    def parameters(self) -> dict[str, str]:
        return self._parameters

    def __str__(self) -> str:
        return self._str

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, MediaType):
            return NotImplemented
        return (self.type, self.subtype, self.parameters) == (other.type, other.subtype, other.parameters)



"""
Provide error handling for the Graphical User Interface.
"""
from __future__ import annotations

import pickle
from asyncio import CancelledError
from logging import getLogger
from traceback import format_exception
from types import TracebackType
from typing import TypeVar, Generic, ParamSpec

from PyQt6.QtCore import QMetaObject, Qt, Q_ARG, QObject
from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QScrollArea, QFrame

from betty.app import App
from betty.error import UserFacingError
from betty.gui.text import Code, Text
from betty.gui.window import BettyMainWindow
from betty.locale import Str, Localizable

T = TypeVar('T')
P = ParamSpec('P')

BaseExceptionT = TypeVar('BaseExceptionT', bound=BaseException)


class ExceptionCatcher(Generic[P, T]):
    """
    Catch any exception and show an error window instead.
    """

    _SUPPRESS_EXCEPTION_TYPES = (
        CancelledError,
    )

    def __init__(
        self,
        parent: QObject,
        *,
        close_parent: bool = False,
    ):
        self._parent = parent
        self._close_parent = close_parent

    def __enter__(self) -> None:
        pass

    def __exit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> bool | None:
        return self._catch(exc_type, exc_val)

    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> bool | None:
        return self._catch(exc_type, exc_val)

    def _catch(self, exception_type: type[BaseExceptionT] | None, exception: BaseExceptionT | None) -> bool | None:
        from betty.gui import BettyApplication

        if exception_type is None or exception is None:
            return None

        if isinstance(exception, self._SUPPRESS_EXCEPTION_TYPES):
            return None

        if isinstance(exception, UserFacingError):
            QMetaObject.invokeMethod(
                BettyApplication.instance(),
                '_show_user_facing_error',
                Qt.ConnectionType.QueuedConnection,
                Q_ARG(type, exception_type),
                Q_ARG(bytes, pickle.dumps(exception)),
                Q_ARG(QObject, self._parent),
                Q_ARG(bool, self._close_parent),
            )
        else:
            getLogger(__name__).exception(exception)
            QMetaObject.invokeMethod(
                BettyApplication.instance(),
                '_show_unexpected_exception',
                Qt.ConnectionType.QueuedConnection,
                Q_ARG(type, exception_type),
                Q_ARG(str, str(exception)),
                Q_ARG(str, ''.join(format_exception(exception))),
                Q_ARG(QObject, self._parent),
                Q_ARG(bool, self._close_parent),
            )
        return True


class Error(BettyMainWindow):
    window_height = 300
    window_width = 500

    def __init__(
        self,
        app: App,
        message: Localizable,
        *,
        parent: QObject,
        close_parent: bool = False,
    ):
        super().__init__(app, parent=parent)
        self._message_localizable = message
        if close_parent and not isinstance(parent, QWidget):
            raise ValueError('If `close_parent` is true, `parent` must be `QWidget`.')
        self._close_parent = close_parent
        self.setWindowModality(Qt.WindowModality.WindowModal)

        central_widget = QWidget()
        self._central_layout = QVBoxLayout()
        central_widget.setLayout(self._central_layout)
        self.setCentralWidget(central_widget)

        self._message = Text()
        self._central_layout.addWidget(self._message)

        self._controls = QHBoxLayout()
        self._central_layout.addLayout(self._controls)

        self._dismiss = QPushButton()
        self._dismiss.released.connect(self.close)
        self._controls.addWidget(self._dismiss)

    @property
    def window_title(self) -> Localizable:
        return Str.plain('{error} - Betty', error=Str._('Error'))

    def _set_translatables(self) -> None:
        super()._set_translatables()
        self._message.setText(self._message_localizable.localize(self._app.localizer))
        self._dismiss.setText(self._app.localizer._('Close'))

    def closeEvent(self, a0: QCloseEvent | None) -> None:
        if self._close_parent:
            parent = self.parent()
            if isinstance(parent, QWidget):
                parent.close()
        super().closeEvent(a0)


ErrorT = TypeVar('ErrorT', bound=Error)


class ExceptionError(Error):
    def __init__(
        self,
        app: App,
        message: Localizable,
        error_type: type[BaseException],
        *,
        parent: QObject,
        close_parent: bool = False,
    ):
        super().__init__(app, message, parent=parent, close_parent=close_parent)
        self.error_type = error_type


class _UnexpectedExceptionError(ExceptionError):
    def __init__(
        self,
        app: App,
        error_type: type[Exception],
        error_message: str,
        error_traceback: str,
        *,
        parent: QObject,
        close_parent: bool = False,
    ):
        super().__init__(
            app,
            Str._(
                'An unexpected error occurred and Betty could not complete the task. Please <a href="{report_url}">report this problem</a> and include the following details, so the team behind Betty can address it.',
                report_url='https://github.com/bartfeenstra/betty/issues',
            ),
            error_type,
            parent=parent,
            close_parent=close_parent,
        )

        if error_message:
            self._exception_message = Code(error_message)
            self._central_layout.addWidget(self._exception_message)

        self._exception_details = QScrollArea()
        self._exception_details.setFrameShape(QFrame.Shape.NoFrame)
        self._exception_details.setWidget(Code(error_traceback + error_traceback + error_traceback + error_traceback))
        self._exception_details.setWidgetResizable(True)
        self._central_layout.addWidget(self._exception_details)



"""
Manage windows.
"""
from __future__ import annotations

from os import path

from PyQt6.QtCore import QObject
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QApplication, QMainWindow

from betty.app import App
from betty.gui.locale import LocalizedObject
from betty.locale import Localizable


class BettyMainWindow(LocalizedObject, QMainWindow):
    window_width = 800
    window_height = 600

    def __init__(
        self,
        app: App,
        *,
        parent: QObject | None = None,
    ):
        super().__init__(app, parent)
        self.resize(self.window_width, self.window_height)
        self.setWindowIcon(QIcon(path.join(path.dirname(__file__), 'assets', 'public', 'static', 'betty-512x512.png')))
        geometry = self.frameGeometry()
        screen = QApplication.primaryScreen()
        assert screen is not None
        geometry.moveCenter(screen.availableGeometry().center())
        self.move(geometry.topLeft())

    def _set_translatables(self) -> None:
        self.setWindowTitle(f'{self.window_title.localize(self._app.localizer)} - Betty')

    @property
    def window_title(self) -> Localizable:
        raise NotImplementedError(repr(self))



"""
Provide file system utilities.
"""
from __future__ import annotations

import asyncio
import hashlib
import os
from collections import deque
from contextlib import suppress
from os.path import getmtime
from pathlib import Path
from shutil import copy2
from types import TracebackType
from typing import AsyncIterable, AsyncContextManager, Sequence

import aiofiles
from aiofiles.os import makedirs
from aiofiles.threadpool.text import AsyncTextIOWrapper

from betty import _ROOT_DIRECTORY_PATH

ROOT_DIRECTORY_PATH = _ROOT_DIRECTORY_PATH


ASSETS_DIRECTORY_PATH = ROOT_DIRECTORY_PATH / 'betty' / 'assets'


HOME_DIRECTORY_PATH = Path.home() / '.betty'


CACHE_DIRECTORY_PATH = HOME_DIRECTORY_PATH / 'cache'


async def iterfiles(path: Path) -> AsyncIterable[Path]:
    """
    Recursively iterate over any files found in a directory.
    """
    for dir_path, _, filenames in os.walk(str(path)):
        for filename in filenames:
            yield Path(dir_path) / filename


def hashfile(path: Path) -> str:
    """
    Get a hash for a file.

    This function relies on the file path and last modified time for uniqueness.
    File contents are ignored.
    """
    return hashlib.md5(':'.join([str(getmtime(path)), str(path)]).encode('utf-8')).hexdigest()


class _Open:
    def __init__(self, fs: FileSystem, file_paths: tuple[Path, ...]):
        self._fs = fs
        self._file_paths = file_paths
        self._file: AsyncContextManager[AsyncTextIOWrapper] | None = None

    async def __aenter__(self) -> AsyncTextIOWrapper:
        for file_path in map(Path, self._file_paths):
            for fs_path, fs_encoding in self._fs._paths:
                with suppress(FileNotFoundError):
                    self._file = aiofiles.open(fs_path / file_path, encoding=fs_encoding)
                    return await self._file.__aenter__()
        raise FileNotFoundError

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: TracebackType | None) -> None:
        if self._file is not None:
            await self._file.__aexit__(None, None, None)


class FileSystem:
    def __init__(self, *paths: tuple[Path, str | None]):
        self._paths = deque(paths)

    def __len__(self) -> int:
        return len(self._paths)

    @property
    def paths(self) -> Sequence[tuple[Path, str | None]]:
        return list(self._paths)

    def prepend(self, path: Path, fs_encoding: str | None = None) -> None:
        self._paths.appendleft((path, fs_encoding))

    def clear(self) -> None:
        self._paths.clear()

    def open(self, *file_paths: Path) -> _Open:
        return _Open(self, file_paths)

    async def copy2(self, source_path: Path, destination_path: Path) -> Path:
        for fs_path, _ in self._paths:
            with suppress(FileNotFoundError):
                await asyncio.to_thread(copy2, fs_path / source_path, destination_path)
                return destination_path
        tried_paths = [str(fs_path / source_path) for fs_path, _ in self._paths]
        raise FileNotFoundError('Could not find any of %s.' % ', '.join(tried_paths))

    async def copytree(self, source_path: Path, destination_path: Path) -> AsyncIterable[Path]:
        file_destination_paths = set()
        for fs_path, _ in self._paths:
            async for file_source_path in iterfiles(fs_path / source_path):
                file_destination_path = destination_path / file_source_path.relative_to(fs_path / source_path)
                if file_destination_path not in file_destination_paths:
                    file_destination_paths.add(file_destination_path)
                    await makedirs(file_destination_path.parent, exist_ok=True)
                    await asyncio.to_thread(copy2, file_source_path, file_destination_path)
                    yield file_destination_path
